# coding=utf-8
"""
WARNING!!!
This module is only meant to be used to ease things with the deleters script in files/deleters. Do not use it anywhere
else.

WARNING!!
Some hacks for easily retrieving paths. Main classes in this module are as given
below:
    DirectoryPaths: Contains some classmethods for directory related operations. These
        classmethods are reimplemented in the Directory class.
    FilePaths: Classmethods for file operations.
    FileObj: A class for accessing files.
    Directory: Base class for folders. Contains useful functions for getting files from
        folders, filtering files etc.
    RootDirectory: This class is useful to assign paths as per a specified as specified
        root directory. It inherits from FilePaths. The _init_ requires a directory path.
    UserDirectory: This class, when instantiated, returns the user home directory,
        pictures.
    makeStringPathSafe: This function will make a string path safe. It searches
        for non alphanumeric characters and replaces them either with nothing or with a specified
        character.
    removeUnsafeChar: This is just a tuple containing potentially unsafe characters. This
    might be useful if I am intereted in making things pathsafe without excluding foreign
    languages like russian chinese etc. In this case, I can just search for thes characters
    in a string.
"""

from __future__ import print_function, division
import os
import warnings
import shutil
import time
import datetime
import platform
import logging
import string
import tempfile


try:
    from .parsedatetime import Calendar
    from .send2trash import send2trash
except ModuleNotFoundError:
    from parsedatetime import Calendar
    from send2trash import send2trash



logger = logging.getLogger("__main__")

string_types__=(str,)

class LoopException(Exception):
    """
    Use this to jump loops.
    """
    pass

def _searchWord(sources, words, includeLineInResults=False, matchCase=True, matchExact=False,
                extInclude=[], extExclude=[], fileNameStrInclude=[], fileNameStrExclude=[],
                fileNameMatchExact=False, dirNameStrInclude=[], dirNameStrExclude=[],
                dirNameMatchExact=False, useExcludeDefaultExts=True, useIncludeDefaultExts=False,
                warnErrors=True, binaryCheck=True, printIgnoredBinaryFiles=False):
    """
    Search for a single word or multiple words in a group of files or directories. Default
    option is to return the line number on which the word occurs. If the includeLineInResults
    is set to True, then the line will be included as well.

    :param sources: This can be a list of files or directories. If directories are provided,
        then the script will search the files inside those directories recursively.
    :param words: A single word or a list of words.
    :param includeLineInResults: Include the line on which the word's occur in the results.
        Default is False.
    :param matchCase: Case sensitive.
    :param matchExact: Word(s) should be exact matches.
    :param extInclude: File extensions to include.
    :param extExclude: File extensions to exclude.
    :param fileNameStrInclude: Include file names containing these strings.
    :param fileNameStrExclude: Exclude file names containing these string.
    :param fileNameMatchExact: The fileName should be exact.
    :param dirNameStrInclude: Include directory names containing these strings.
    :param dirNameStrExclude: Exclude directory names containing these strings.
    :param dirNameMatchExact: The directory name match should be exact.
    :param useExcludeDefaultExts: Some default files such as .exe, .png etc have been
        already excluded. Set this to False to consider them as well.
    :param useIncludeDefaultExts:Some default files such as .txt, .log etc have been
        already included. Set this to False to ignore those defaults.
    :param warnErrors: Raising warnings in cases of errors.
    :return:
    """


    ignoredBinaryFiles = []

    extIncludeDefault = ['.py', '.txt', 'xls', '.xlsm', '.doc', '.htm', '.html', '.rtf', '.docx', '.bat', '.csv',
                         '.log']
    extExcludeDefault = ['.exe', '.png', '.bmp', '.jpg', '.gh', '.3dm', '.hdr']

    # If an input is specified as a string convert into a list with one element.
    convertToList = lambda value: [value] if isinstance(value, str) else value

    extInclude, extExclude, fileNameStrExclude, fileNameStrInclude = map(convertToList, (
    extInclude, extExclude, fileNameStrExclude, fileNameStrInclude))
    dirNameStrExclude, dirNameStrInclude = map(convertToList, (dirNameStrExclude, dirNameStrInclude))
    sources = convertToList(sources)

    if useExcludeDefaultExts:
        extExclude += extExcludeDefault

    if useIncludeDefaultExts:
        extInclude += extIncludeDefault

    # start a dictionary for storing all the words and results.
    if not isinstance(words, (list, tuple, set)):
        wordDict = {words: []}
    else:
        wordDict = dict.fromkeys(words, [])

    for source in sources:
        if os.path.isdir(source):
            for dirs, subs, files in os.walk(source):

                dirsNormalized = os.path.normpath(dirs).lower()
                dirsSplit = dirsNormalized.split(os.sep)
                if dirNameMatchExact:
                    dirName = dirsSplit
                else:
                    dirName = dirs.lower()

                if dirNameStrInclude:
                    try:
                        for includeStrings in map(lambda x: x.lower(), dirNameStrInclude):
                            if includeStrings not in dirName:
                                raise LoopException(
                                    "Custom exception to get out of the loop in case include lists aren't matched.")
                    except LoopException:
                        continue

                if dirNameStrExclude:
                    try:
                        for excludeStrings in map(lambda x: x.lower(), dirNameStrExclude):
                            if excludeStrings in dirName:
                                raise LoopException(
                                    "Custom exception to get out of the loop in case include lists aren't matched.")
                    except LoopException:
                        continue

                for fileName in files:

                    fileNameCurrent = os.path.join(dirs, fileName)

                    # Check for extensions.
                    ext = os.path.splitext(fileNameCurrent)[1]
                    if extInclude:
                        if not ext in extInclude:
                            continue
                    if extExclude:
                        if ext in extExclude:
                            continue

                    # Check for filename strings.
                    directoryName, fileNameOnly = os.path.split(fileName)

                    if fileNameStrExclude:
                        try:
                            if not fileNameMatchExact:
                                if fileNameOnly.lower() in map(lambda x: x.lower(), fileNameStrExclude):
                                    continue
                            else:
                                for string in map(lambda x: x.lower(), fileNameStrExclude):
                                    if fileNameOnly.lower() == string:
                                        raise LoopException
                        except LoopException:
                            continue

                    if fileNameStrInclude:
                        try:
                            if not fileNameMatchExact:
                                if not fileNameOnly.lower() in map(lambda x: x.lower(), fileNameStrInclude):
                                    continue
                            else:
                                for string in map(lambda x: x.lower(), fileNameStrInclude):
                                    if fileNameOnly.lower() != string:
                                        raise LoopException
                        except LoopException:
                            continue

                    try:
                        with open(fileNameCurrent)as currentFile:

                            if binaryCheck:
                                try:
                                    currentFile.read(1024)
                                except UnicodeDecodeError:
                                    raise LoopException

                            for idx, lines in enumerate(currentFile):
                                for word in wordDict:
                                    linesTemp = lines
                                    originalLine = lines
                                    originalWord = word
                                    if not matchCase:
                                        linesTemp = lines.lower()
                                        word = word.lower()
                                    if matchExact:
                                        linesTemp = lines.split()

                                    if word in linesTemp:
                                        if includeLineInResults:
                                            wordDict[originalWord].append((fileNameCurrent, idx, originalLine))
                                        else:
                                            wordDict[originalWord].append((fileNameCurrent, idx + 1))
                                pass
                    except IOError:
                        if warnErrors:
                            msg = "IOError reported for {}".format(fileNameCurrent)
                            warnings.warn(msg)
                        else:
                            pass
                    except LoopException:
                        ignoredBinaryFiles.append(fileNameCurrent)

    if ignoredBinaryFiles and printIgnoredBinaryFiles:
        msg = "\n".join(ignoredBinaryFiles)
        msg = "The following files were found to be binary and ignored:\n%s" % msg
        warnings.warn(msg)

    return wordDict

class FilePaths(object):
    r"""The FilePaths class contains a lot of class methods that can be really useful
    for path related operations. Following is a description of methods.

        fileNameOnly: Return only the name of the file with path or extension.
               For example fileNameOnly(c:\users\sarith\jasmine.jpg) will return jasmine.

        changeExtension: Change the extension for path. Like make c:\users\sarith\x.jpg to
            c:\users\sarith\x.bmp

    """

    @classmethod
    def pathToDifferentDirectory(cls, filePath, directory):
        r"""
        Generate a new path for the current file in a directory. This can be useful in creating a pathname for
        writing. For example: c:\users\sarith\desktop\pp.rad to d:\stuff\pp.rad
        :param filePath:
        :param directory:
        :return:
        """
        oldDirectory, fileName = os.path.split(filePath)
        assert oldDirectory != directory, 'The old directory %s and the directory to move to %s are the same.' % (
            oldDirectory, directory)
        newPath = os.path.join(directory, fileName)
        return newPath

    @classmethod
    def moveToDifferentDirectory(cls, filePath, directory):
        r"""
        Copy the current file to a different directory. This can be useful in creating a pathname for
        writing. For example: c:\users\sarith\desktop\pp.rad to d:\stuff\pp.rad
        :param filePath:
        :param directory:
        :return:
        """
        newPath = cls.pathToDifferentDirectory(filePath, directory)
        shutil.copy(filePath, newPath)

    @classmethod
    def addSuffixToName(cls, filePath, suffix):
        r"""
        Add a suffix to the filename. For example, make c:\users\sarith\apple.txt to c:\users\sarith\appleBee.txt
        :param filePath: File path.
        :param suffix: The suffix to be added.
        :return:
        """
        path, fileName = os.path.split(filePath)
        name, extension = os.path.splitext(fileName)
        name = name + suffix + extension
        return os.path.join(path, name)

    @classmethod
    def addPrefixToName(cls, filePath, prefix):
        r"""
        Add a prefix to the filename. For example, make c:\users\sarith\apple.txt to c:\users\sarith\beeapple.txt
        :param filePath:
        :param prefix:
        :return:
        """
        path, fileName = os.path.split(filePath)
        fileName = prefix + fileName
        return os.path.join(path, fileName)

    @classmethod
    def fileNameOnly(cls, filePath):
        r"""Return only the name of the file with path or extension.
           For example fileNameOnly(c:\users\sarith\jasmine.jpg) will return jasmine.
        """
        filePathOnly = os.path.split(filePath)[1]
        fileNameOnly = os.path.splitext(filePathOnly)[0]
        return fileNameOnly

    @classmethod
    def changeExtension(cls, filePath, newExtension=None):
        r"""Change the extension for path. Like make c:\users\sarith\x.jpg to c:\users\sarith\x.bmp"""

        filePathOnly = os.path.splitext(filePath)[0]
        if newExtension:
            assert isinstance(newExtension, string_types__), 'The extension should be a string'
            newExtension = newExtension.replace('.', '')
            return filePathOnly + '.' + newExtension
        else:
            return filePathOnly

class FileObj(object):
    r"""Class for a filepath.

        Usage:
            1. Simple implementation
            x = FileObj(r"c:\something.txt")

            #relative path
            x.relPath

            #absolute path
            x.absPath

            2. Assertiontesting with variable name. Note that usually, the path will
                be specified as a variable and not as a string.

            x=FileObj(somefile,"somefile")

        The following methods and properties are implemented.
        nameOnly: return only he name of the file.
        abspath: absolute path.
        relPath: relative path.
        path: return the path.
        changeExtension: strip the extension or replace with a new one if provided.
        searchWord: look for specific words within the file.
        sizeBytes: size in bytes.
        sizeKBytes: in kilo bytes.
        sizeMBytes: in MB
    """

    def __init__(self, fileName,fileVariableName=None):
        """
        :param directoryName: FilePath of the file.
        :param fileVariableName: Name of the variable to which the path is assigned. This
            is optional but pretty helpful in case of assertion testing.
        """
        self._fileVariableName=fileVariableName
        self.fileName = fileName

    def pathToDifferentDirectory(self, directory):
        r"""
        Generate a new path for the current file in a directory. This can be useful in creating a pathname for
        writing. For example: c:\users\sarith\desktop\pp.rad to d:\stuff\pp.rad
        :param directory:
        :return:
        """
        return FilePaths.pathToDifferentDirectory(self.fileName, directory)

    def moveToDifferentDirectory(self, directory):
        r"""
        Copy the current file to a different directory. This can be useful in creating a pathname for
        writing. For example: c:\users\sarith\desktop\pp.rad to d:\stuff\pp.rad
        :param directory:
        :return:
        """
        return FilePaths.moveToDifferentDirectory(self.fileName, directory)

    def addSuffixToName(self, suffix):
        r"""
        Add a suffix to the filename. For example, make c:\users\sarith\apple.txt to c:\users\sarith\appleBee.txt
        :param suffix: The suffix to be added.
        :return:
        """
        return FilePaths.addSuffixToName(self.fileName, suffix)

    def addPrefixToName(self, prefix):
        r"""
        Add a prefix to the filename. For example, make c:\users\sarith\apple.txt to c:\users\sarith\beeapple.txt
        :param prefix:
        :return:
        """

        return FilePaths.addSuffixToName(self.fileName, prefix)

    @property
    def fileName(self):
        return self._fileName

    @fileName.setter
    def fileName(self, value):
        r"""

        :param value: pathname
        :return:
        """
        if self._fileVariableName:
            assert os.path.isfile(value), 'The path for %s(%s) is not a valid file. The current working ' \
                                          'directory is %s' % (self._fileVariableName,value, os.getcwd())
        else:
            assert os.path.isfile(value), '%s is not a valid file. The current working ' \
                                          'directory is %s' % (value, os.getcwd())
        self._fileName = value

    @property
    def nameOnly(self):
        r"""Return only the name of the file. ie 'sarith' if path is c:\users\sarith.txt"""

        nameOnly = FilePaths.fileNameOnly(self.fileName)
        return nameOnly

    @property
    def absPath(self):
        return os.path.abspath(self.fileName)

    @property
    def relPath(self):
        return os.path.relpath(self.fileName)

    def __repr__(self):
        return str(self.fileName)

    def __str__(self):
        return str(self.fileName)

    @property
    def path(self):
        return str(self.fileName)

    def changeExtension(self, newExtension=None):
        r"""
        Strip the extension from the filepath. If a newExtension is provided then add that
        to the file path. This can be useful in Radiance type enviroments where a x.hdr
        is being converted x.tif etc.
        :param newExtension:
        :return:
        """
        newPath = FilePaths.changeExtension(self.fileName, newExtension)
        return newPath

    def searchWord(self, words, includeLineInResults=False, matchCase=True,
                   matchExact=False,
                   extInclude=[], extExclude=[], fileNameStrInclude=[],
                   fileNameStrExclude=[],
                   fileNameMatchExact=False, dirNameStrInclude=[], dirNameStrExclude=[],
                   dirNameMatchExact=False, useExcludeDefaultExts=True,
                   useIncludeDefaultExts=False,
                   warnErrors=True, binaryCheck=True, printIgnoredBinaryFiles=False):
        r"""
        Search for a single word or multiple words in this fileName. Default
        option is to return the line number on which the word occurs. If the includeLineInResults
        is set to True, then the line will be included as well.

        :param sources: This can be a list of files or directories. If directories are provided,
            then the script will search the files inside those directories recursively.
        :param words: A single word or a list of words.
        :param includeLineInResults: Include the line on which the word's occur in the results.
            Default is False.
        :param matchCase: Case sensitive.
        :param matchExact: Word(s) should be exact matches.
        :param extInclude: File extensions to include.
        :param extExclude: File extensions to exclude.
        :param fileNameStrInclude: Include file names containing these strings.
        :param fileNameStrExclude: Exclude file names containing these string.
        :param fileNameMatchExact: The fileName should be exact.
        :param dirNameStrInclude: Include directory names containing these strings.
        :param dirNameStrExclude: Exclude directory names containing these strings.
        :param dirNameMatchExact: The directory name match should be exact.
        :param useExcludeDefaultExts: Some default files such as .exe, .png etc have been
            already excluded. Set this to False to consider them as well.
        :param useIncludeDefaultExts:Some default files such as .txt, .log etc have been
            already included. Set this to False to ignore those defaults.
        :param warnErrors: Raising warnings in cases of errors.
        :return
        """

        return _searchWord(self.fileName, words,
                           includeLineInResults=includeLineInResults,
                           matchCase=matchCase, matchExact=matchExact,
                           extInclude=extInclude,
                           extExclude=extExclude, fileNameStrInclude=fileNameStrInclude,
                           fileNameStrExclude=fileNameStrExclude,
                           fileNameMatchExact=fileNameMatchExact,
                           dirNameStrInclude=dirNameStrInclude,
                           dirNameStrExclude=dirNameStrExclude,
                           dirNameMatchExact=dirNameMatchExact,
                           useExcludeDefaultExts=useExcludeDefaultExts,
                           useIncludeDefaultExts=useIncludeDefaultExts,
                           warnErrors=warnErrors,
                           binaryCheck=binaryCheck,
                           printIgnoredBinaryFiles=printIgnoredBinaryFiles)

    @property
    def sizeBytes(self):
        r"""Size in bytes"""
        return os.stat(self.fileName).st_size

    @property
    def sizeKBytes(self):
        r"""Size in kilobytes rounded off to single decimal place."""
        return round(os.stat(self.fileName).st_size / 1024, 1)

    @property
    def sizeMBytes(self):
        r"""Size in megabytes rounded off to four decimal places."""

        return round(os.stat(self.fileName).st_size / (1024 * 1024), 4)

    @property
    def timeModified(self):
        r"""
        Return the last modified time in datetime.datetime format.
        @:rtype: datetime.datetime

        :return:
        """
        return datetime.datetime.fromtimestamp(os.path.getmtime(self.absPath))

    @property
    def timeCreated(self):
        r"""
        Return the last modified time in datetime.datetime format.

        Try to get the date that a file was created, falling back to when it was
        last modified if that isn't possible.
        See http://stackoverflow.com/a/39501288/1709587 for explanation.

        @:rtype: datetime.datetime
        """
        if platform.system() == 'Windows':
            return datetime.datetime.fromtimestamp(os.path.getctime(self.absPath))
        else:
            stat = datetime.datetime.fromtimestamp(os.stat(self.absPath))
            try:
                return datetime.datetime.fromtimestamp(stat.st_birthtime)
            except AttributeError:
                # We're probably on Linux. No easy way to get creation dates here,
                # so we'll settle for when its content was last modified.
                return datetime.datetime.fromtimestamp(stat.st_mtime)

class DirectoryPaths(object):
    r"""The FilePaths class contains a lot of class methods that can be really useful
    for path related operations. Following is a description of methods.

            filesInDirRelative: Return relative path names from a directory. For example, for a
            directory 'room" containing 'floor.rad', 'ground.rad' the result will be
            ['floor.rad,'ground.rad']

        filesInDirAbsolute: Return absolute path names from a directory. For example, for
            a directory 'home/room" containing 'floor.rad', 'ground.rad' the result will
            be ['home/room/floor.rad,'home/room/ground.rad']

        fileNamePathDictionary:  Return a dictionary containing keys as fileNames (only)
            and values as filePaths. Absolute paths will be returned by default, though
            this can be changed. For a folder "/home/room/geometry" containing 'floor.rad',
            'ceiling.rad', 'walls.rad' the output will be
            {'floor':'/home/room/geometry/floor.rad', 'ceiling':'/home/room/geometry/ceiling.rad',
            'walls':'/home/room/geometry/walls.rad'}

        findDuplicateNames: Search for duplicates names (only) in a directory. For example
            if a directory room contains [room.rad,room.dwg, room.stl, floor.rad,floor.stl,
            ground.rad], then the output will be {'room':['.stl','.rad','.dwg'],
            'floor':['.rad','.stl'] If the returnExtensions option is set to False, then
            only numbers will be returned.


    """

    @classmethod
    def filesInDirRelative(cls, dirName, filesOnly=True, dirsOnly=False, filterExtensionIncl=None,
                           filterExtensionExcl=None,
                           filterLettersExclude=None, filterLettersInclude=None,
                           filterEndsWith=None, filterStartsWith=None, letterSearchCaseSensitive=True,
                           sortByName=True,filterExtensionsCaseSensitive=False):
        r"""
        Return relative path names from a directory. For example, for a directory 'room"
        containing 'floor.rad', 'ground.rad' the result will be ['floor.rad,'ground.rad']

        :param dirName: directory of interest. Required.
        :param filesOnly: look for files only.
        :param dirsOnly: look for directories only.
        :param filterExtensionIncl: Include these extensions.
        :param filterExtensionExcl: Exclude these extensions.
        :param filterLettersExclude: Look for entries that do not contain this string or
            list of strings.
        :param filterLettersInclude: Look for entries that contain this string or list of
            strings.
        :param filterEndsWith: Look for entries that end with this string. Note that the
            extension should be included with the string. ..i.e. "oom.rad"
        :param filterStartsWith: Look for entries that start with this string. The searching
            for this begins at the file level only and not the directory level.
        :param letterSearchCaseSensitive: If strings are being searched then setting this to
            False makes the search case insensitive.
        :param sortByName: If set to True, then sort the filenames by alphabetical order.
        :param filterExtensionsCaseSensitive: If set to True then filterLettersInclude and filterLettersExclude options will
            be considered on a case-sensitive basis.
        :return: List of files/directories in folder dirName.
        """
        tempString = "Only one of those options should be set."

        putStringInList = lambda x: [x] if isinstance(x, string_types__) else list(x) if x else None

        assert not (filesOnly and dirsOnly), 'Both filesOnly and dirsOnly have been set ' \
                                             'to True.%s' % tempString

        assert not (filterExtensionExcl and filterExtensionIncl), \
            'Both filterExtensionExcl and filterExtensionIncl have been ' \
            'set. %s' % tempString

        assert not (filterLettersExclude and filterLettersInclude), \
            'Both filterLettersExclude and filterLettersInclude have been ' \
            'set. %s' % tempString



        filterEndsWith = putStringInList(filterEndsWith)
        filterStartsWith = putStringInList(filterStartsWith)
        filterExtensionExcl = putStringInList(filterExtensionExcl)
        filterExtensionIncl = putStringInList(filterExtensionIncl)
        filterLettersExclude=putStringInList(filterLettersExclude)
        filterLettersInclude=putStringInList(filterLettersInclude)

        if filterEndsWith:
            logger.debug("Filter criteria| Ends with: %s"%','.join(filterEndsWith))
        if filterStartsWith:
            logger.debug("Filter criteria| Starts with: %s"%','.join(filterStartsWith))
        if filterExtensionExcl:
            logger.debug("Filter criteria| Extensions to exclude: %s" % ','.join(filterExtensionExcl))
        if filterExtensionIncl:
            logger.debug("Filter criteria| Extensions to include: %s" % ','.join(filterExtensionIncl))
        if filterLettersExclude:
            logger.debug("Filter criteria| Letter to exclude: %s" % ','.join(filterLettersExclude))
        if filterLettersInclude:
            logger.debug("Filter criteria| Letter to include: %s" % ','.join(filterLettersInclude))

        if filterExtensionExcl or filterExtensionIncl:
            logger.debug("The filtering of extensions is%scase-sensitive"%(" " if filterExtensionsCaseSensitive else " not "))

        logger.debug("The filtering of letters is%scase-sensitive"%(" " if letterSearchCaseSensitive else " not "))

        pathNames = os.listdir(dirName)

        logger.debug("Total number of files and folders in %s are %s"%(dirName,len(pathNames)))

        if filesOnly:
            pathNames = [pathName for pathName in pathNames if os.path.isfile(os.path.join(dirName, pathName))]
            logger.debug("The number of files alone is %s"%(len(pathNames)))

        if dirsOnly:
            pathNames = [pathName for pathName in pathNames if os.path.isdir(os.path.join(dirName, pathName))]
            logger.debug("The number of folders alone is %s" % (len(pathNames)))

        if filterExtensionExcl and not dirsOnly:
            filterExtensionExcl = ['.' + x if not x.startswith('.') else x for x in filterExtensionExcl]
            if filterExtensionsCaseSensitive:
                pathNames = [pathName for pathName in pathNames if os.path.splitext(pathName)[1] not in filterExtensionExcl]
            else:
                pathNames = [pathName for pathName in pathNames if os.path.splitext(pathName)[1].lower() not in list(map(str.lower, filterExtensionExcl))]

            logger.debug("Number of after filtering as per extensions are %s"%(len(pathNames)))

        if filterExtensionIncl and not dirsOnly:
            filterExtensionIncl = ['.' + x if not x.startswith('.') else x for x in filterExtensionIncl]
            if filterExtensionsCaseSensitive:
                pathNames = [pathName for pathName in pathNames if os.path.splitext(pathName)[1] in filterExtensionIncl]
            else:
                pathNames = [pathName for pathName in pathNames if os.path.splitext(pathName)[1].lower() in list(map(str.lower,filterExtensionIncl))]

            logger.debug("Number of files after filtering as per extensions are %s" % (len(pathNames)))

        if filterEndsWith:
            newPathNames = []
            for stringVal in filterEndsWith:
                if letterSearchCaseSensitive:
                    newPathNames.extend([pathName for pathName in pathNames if pathName.endswith(stringVal)])
                else:
                    newPathNames.extend([pathName for pathName in pathNames if pathName.lower().endswith(stringVal.lower())])
            pathNames = list(newPathNames)

            logger.debug("Number of files/folders after filtering as per ending letters are %s" % (len(pathNames)))

        if filterStartsWith:
            newPathNames = []
            for stringVal in filterStartsWith:
                if letterSearchCaseSensitive:
                    newPathNames.extend([pathName for pathName in pathNames if pathName.startswith(stringVal)])
                else:
                    newPathNames.extend([pathName for pathName in pathNames if pathName.lower().startswith(stringVal.lower())])

            pathNames = list(newPathNames)
            logger.debug("Number of files/folders after filtering as per starting letters are %s" % (len(pathNames)))

        if filterLettersExclude:
            newPathNames=[]
            rejectList=[]
            if letterSearchCaseSensitive:
                for stringVal in filterLettersExclude:

                    currentSearchList=list(newPathNames) or list(pathNames)
                    newPathNames=[]
                    for pathName in currentSearchList:
                        if (not (stringVal in pathName))  and (not (pathName in rejectList)):
                            newPathNames.append(pathName)
                        else:
                            rejectList.append(pathName)
            else:
                for stringVal in filterLettersExclude:
                    currentSearchList = list(newPathNames) or list(pathNames)
                    newPathNames = []

                    for pathName in currentSearchList:
                        if (not (stringVal.lower() in pathName.lower()))  and (not (pathName in rejectList)):
                            newPathNames.append(pathName)
                        else:
                            rejectList.append(pathName)

            pathNames = list(newPathNames)
            logger.debug("Number of files/folders after filtering as per exclusion of letters are %s" % (len(pathNames)))
        if filterLettersInclude:
            newPathNames=[]
            if letterSearchCaseSensitive:
                for stringVal in filterLettersInclude:
                    for pathName in pathNames:
                        if (stringVal in pathName) and not (pathName in newPathNames):
                            newPathNames.append(pathName)
            else:
                for stringVal in filterLettersInclude:
                    for pathName in pathNames:
                        if  (stringVal.lower() in pathName.lower()) and not (pathName in newPathNames):
                            newPathNames.append(pathName)

            pathNames = list(newPathNames)
            logger.debug("Number of files/folders after filtering as per inclusion of letters are %s" % (len(pathNames)))

        if sortByName:
            pathNames=list(sorted(pathNames))

        return pathNames

    @classmethod
    def filesInDirAbsolute(cls, dirName, filesOnly=True, dirsOnly=False, filterExtensionIncl=None,
                           filterExtensionExcl=None,
                           filterLettersExclude=None, filterLettersInclude=None,
                           filterEndsWith=None, filterStartsWith=None, letterSearchCaseSensitive=True,
                           sortByName=True,filterExtensionsCaseSensitive=False):
        r"""
        Return absolute path names from a directory. For example, for a directory 'home/room"
        containing 'floor.rad', 'ground.rad' the result will be ['home/room/floor.rad,'home/room/ground.rad']

        :param dirName: directory of interest. Required.
        :param filesOnly: look for files only.
        :param dirsOnly: look for directories only.
        :param filterExtensionIncl: Include these extensions.
        :param filterExtensionExcl: Exclude these extensions.
        :param filterLettersExclude: Look for entries that do not contain this string or
            list of strings.
        :param filterLettersInclude: Look for entries that contain this string or list of
            strings.
        :param filterEndsWith: Look for entries that end with this string. Note that the
            extension should be included with the string. ..i.e. "oom.rad"
        :param filterStartsWith: Look for entries that start with this string. The searching
            for this begins at the file level only and not the directory level.
        :param letterSearchCaseSensitive: If strings are being searched then setting this to
            False makes the search case insensitive.
        :param filterExtensionsCaseSensitive: If set to True then filterLettersInclude and filterLettersExclude options will
            be considered on a case-sensitive basis.
        :param sortByName: If set to True, then sort the filenames by alphabetical order
        :return: List of files/directories in folder dirName.
        """

        pathNames = cls.filesInDirRelative(dirName, filesOnly, dirsOnly, filterExtensionIncl, filterExtensionExcl,
                                           filterLettersExclude, filterLettersInclude,
                                           filterEndsWith, filterStartsWith, letterSearchCaseSensitive,
                                           sortByName=sortByName,filterExtensionsCaseSensitive=filterExtensionsCaseSensitive)

        return [os.path.join(dirName, fileName) for fileName in pathNames]

    @classmethod
    def fileNamePathDictionary(cls, dirName, returnAbsPaths=True, filterExtensionIncl=None,
                               filterExtensionExcl=None, filterLettersExclude=None,
                               filterLettersInclude=None, filterEndsWith=None,
                               filterStartsWith=None, letterSearchCaseSensitive=True,filterExtensionsCaseSensitive=False):
        r"""
        Return a dictionary containing keys as fileNames (only) and values as filePaths.
        Absolute paths will be returned by default, though this can be changed.
        For a folder "/home/room/geometry" containing 'floor.rad', 'ceiling.rad', 'walls.rad'
        the output will be {'floor':'/home/room/geometry/floor.rad',
                            'ceiling':'/home/room/geometry/ceiling.rad',
                            'walls':'/home/room/geometry/walls.rad'}

        :param dirName: directory of interest. Required.
        :param filterExtensionIncl: Include these extensions.
        :param filterExtensionExcl: Exclude these extensions.
        :param filterLettersExclude: Look for entries that do not contain this string.
        :param filterLettersInclude: Look for entries that contain this string.
        :param filterEndsWith: Look for entries that end with this string. Note that the
            extension should be included with the string. ..i.e. "oom.rad"
        :param filterStartsWith: Look for entries that start with this string. The searching
            for this begins at the file level only and not the directory level.
        :param letterSearchCaseSensitive: If strings are being searched then setting this to
            False makes the search case insensitive.
        :param filterExtensionsCaseSensitive: If set to True then filterLettersInclude and filterLettersExclude options will
            be considered on a case-sensitive basis.
        :return: List of files/directories in folder dirName.

        """
        filesOnly = True
        dirsOnly = False
        pathNames = cls.filesInDirRelative(dirName, filesOnly, dirsOnly, filterExtensionIncl, filterExtensionExcl,
                                           filterLettersExclude, filterLettersInclude,
                                           filterEndsWith, filterStartsWith, letterSearchCaseSensitive,filterExtensionsCaseSensitive=filterExtensionsCaseSensitive)
        if returnAbsPaths:
            pathNames = [os.path.join(dirName, fileName) for fileName in pathNames]

        fileNames = map(FilePaths.fileNameOnly, pathNames)

        fileDict = dict(zip(fileNames, pathNames))

        assert len(fileDict) == len(pathNames), 'The number of files in the dictionary(%s) ' \
                                                'is not equal to the number of files in ' \
                                                'the directory %s. It is likely that there' \
                                                'are multiple files with same names but ' \
                                                'different extensions in this directory (%s). Use ' \
                                                'the findDuplicateNames function to find out ' \
                                                'what those duplicates are. Then filterExtensionExcl' \
                                                'can be used to exclude those extensions' % (
                                                len(fileDict), len(pathNames), dirName)
        return fileDict

    @classmethod
    def findDuplicateNames(cls, dirName, filesOnly=True, dirsOnly=False,
                           filterExtensionIncl=None, filterExtensionExcl=None,
                           filterLettersExclude=None, filterLettersInclude=None,
                           filterEndsWith=None, filterStartsWith=None,
                           letterSearchCaseSensitive=True, returnExtensions=True,filterExtensionsCaseSensitive=False):
        r"""
        Search for duplicates names (only) in a directory. For example if a directory
        room contains [room.rad,room.dwg, room.stl, floor.rad,floor.stl,ground.rad],
        then the output will be {'room':['.stl','.rad','.dwg'],'floor':['.rad','.stl']
        If the returnExtensions option is set to False, then only numbers will be returned.

        :param dirName: directory of interest. Required.
        :param filesOnly: look for files only.
        :param dirsOnly: look for directories only.
        :param filterExtensionIncl: Include these extensions.
        :param filterExtensionExcl: Exclude these extensions.
        :param filterLettersExclude: Look for entries that do not contain this string.
        :param filterLettersInclude: Look for entries that contain this string.
        :param filterEndsWith: Look for entries that end with this string. Note that the
            extension should be included with the string. ..i.e. "oom.rad"
        :param filterStartsWith: Look for entries that start with this string. The searching
            for this begins at the file level only and not the directory level.
        :param letterSearchCaseSensitive: If strings are being searched then setting this to
            False makes the search case insensitive.
        :param filterExtensionsCaseSensitive: If set to True then filterLettersInclude and filterLettersExclude options will
            be considered on a case-sensitive basis.
        :param returnExtensions: Return extensions of the files containing duplicate names.
        :return: List of files/directories in folder dirName.
        """
        pathNames = cls.filesInDirRelative(dirName, filesOnly, dirsOnly, filterExtensionIncl, filterExtensionExcl,
                                           filterLettersExclude, filterLettersInclude,
                                           filterEndsWith, filterStartsWith, letterSearchCaseSensitive,
                                           filterExtensionsCaseSensitive=filterExtensionsCaseSensitive)
        fileNames = map(FilePaths.fileNameOnly, pathNames)

        fileDict = dict.fromkeys(fileNames, None)

        for idx, name in enumerate(fileNames):
            ext = os.path.splitext(pathNames[idx])[1]
            if fileDict[name] is None:
                fileDict[name] = [ext]
            else:
                fileDict[name].append(ext)

        for key, value in fileDict.items():
            if len(value) == 1:
                del (fileDict[key])
            elif not returnExtensions:
                fileDict[key] = len(value)
        return fileDict

class Directory(object):
    r"""Use this to navigate through files in a directory with filters etc.
    Note that I am not inheriting directoryPaths as I don't want those classmethods in
    DirectoryPaths to be available in Directory.
    Following methods are implemented.
        fileNamesAbsolute: returns absolute fileNames with a lot of bells and whistles.
        fileNamesRelative: returns relative fileNames with a lot of bells and whisltes.
        fileDictionary: Returns a dictionary containing keys as fileNames(only) and
            pathName as values. It is essential that there be only one type of file for
            each name. This method can be useful for iterating through a list of files
            and writing some other files as sideEffects.
        filesDuplicate: Return a dictionary cotnaining duplicate fileNames in a directory.
            This can be used to troubleshoot when the fileDictionary method fails.
        subDirectory: return a new Directory instance that is a subfolder inside the
            current directory. This can be used recursively like
            a.subDirectory('room').subDirectory('geometry').. as long as those directories
            exist.
        searchWord: look for word(s) inside current directory.
        fileInside: return the path for a single file that matches the letters provided.
        fileObjInside: return the FileObj for a single file that matches the letters provided.

        Usage:
            1. Initialize a simple directory:
                x = Directory(r"c:\somedirectory")

                # Return absolute path
                x.absPath

                # Return relative path.
                x.relPath

            2. Initialize a directory while also doing an assertion test.

                #This is useful for testing if the path assigned for a given variable exists
                # and then printing an assertion statement to this effect.

                x = Directory(r"c:\iesDirectory",directoryVariableName="iesData")


    """

    def __init__(self, directoryName, directoryVariableName=None, createPathIfNotPresent=False):
        r"""
        :param directoryName: FilePath of the directory.
        :param directoryVariableName: This is an optional value to be used mostly
            for assertion testing. More info in usage notes.
        :param createPathIfNotPresent: If the directory is not present, then create the entire
            directory structure.

        """
        self._createIfNotPresent=createPathIfNotPresent
        self._directoryVariableName = directoryVariableName
        self.directoryName = directoryName


    @property
    def directoryName(self):
        return self._directoryName

    @directoryName.setter
    def directoryName(self, value):
        if self._directoryVariableName:
            if self._createIfNotPresent:
                try:
                    newPath=os.path.join(value,self._directoryVariableName)
                    os.makedirs(newPath)
                    logger.debug("The path %s did not exist and was created."%newPath)
                except FileExistsError:
                    logger.debug("The path %s exists"%newPath)
            else:
                assert os.path.isdir(value), 'The path for variable %s(%s) is not valid. The current working ' \
                                         'directory is %s' % (self._directoryVariableName,value, os.getcwd())
        else:
            if self._createIfNotPresent:
                try:
                    os.makedirs(value)
                    logger.debug("The path %s did not exist and was created." % value)
                except FileExistsError:
                    logger.debug("The path %s exists" % value)
            else:
                assert os.path.isdir(
                    value), '%s is not a valid directory. The current working ' \
                            'directory is %s' % (value, os.getcwd())
        self._directoryName = value

    @property
    def directoryNameOnly(self):
        """Return only the name of the directory and nothing else."""
        return os.path.split(self.absPath)[-1]

    def fileNamesAbsolute(self, filesOnly=True, dirsOnly=False,
                          filterExtensionIncl=None, filterExtensionExcl=None,
                          filterLettersExclude=None, filterLettersInclude=None,
                          filterEndsWith=None, filterStartsWith=None,
                          letterSearchCaseSensitive=True,sortByName=True,filterExtensionsCaseSensitive=False):
        r"""
        Return absolute path names from a directory. For example, for a directory 'home/room"
        containing 'floor.rad', 'ground.rad' the result will be ['home/room/floor.rad,'home/room/ground.rad']

        :param filesOnly: look for files only.
        :param dirsOnly: look for directories only.
        :param filterExtensionIncl: Include these extensions.
        :param filterExtensionExcl: Exclude these extensions.
        :param filterLettersExclude: Look for entries that do not contain this string.
        :param filterLettersInclude: Look for entries that contain this string.
        :param filterEndsWith: Look for entries that end with this string. Note that the
            extension should be included with the string. ..i.e. "oom.rad"
        :param filterStartsWith: Look for entries that start with this string. The searching
            for this begins at the file level only and not the directory level.
        :param letterSearchCaseSensitive: If strings are being searched then setting this to
            False makes the search case insensitive.
        :param filterExtensionsCaseSensitive: If set to True then filterLettersInclude and filterLettersExclude options will
            be considered on a case-sensitive basis.
        :param sortByName: If set to True, then sort the filenames by alphabetical order
       :return: List of files/directories in folder dirName.


        """
        files = DirectoryPaths.filesInDirAbsolute(self.directoryName, filesOnly=filesOnly,
                                                  dirsOnly=dirsOnly, filterExtensionIncl=filterExtensionIncl,
                                                  filterExtensionExcl=filterExtensionExcl,
                                                  filterLettersExclude=filterLettersExclude,
                                                  filterLettersInclude=filterLettersInclude,
                                                  filterEndsWith=filterEndsWith,
                                                  filterStartsWith=filterStartsWith,
                                                  letterSearchCaseSensitive=letterSearchCaseSensitive,
                                                  sortByName=sortByName,
                                                  filterExtensionsCaseSensitive=filterExtensionsCaseSensitive)
        return files

    def fileNamesRelative(self, filesOnly=True, dirsOnly=False, filterExtensionIncl=None,
                          filterExtensionExcl=None, filterLettersExclude=None,
                          filterLettersInclude=None, filterEndsWith=None,
                          filterStartsWith=None, letterSearchCaseSensitive=True,
                          sortByName=True,filterExtensionsCaseSensitive=False):
        r"""
        Return relative path names from a directory. For example, for a directory 'room"
        containing 'floor.rad', 'ground.rad' the result will be ['floor.rad,'ground.rad']

        :param filesOnly: look for files only.
        :param dirsOnly: look for directories only.
        :param filterExtensionIncl: Include these extensions.
        :param filterExtensionExcl: Exclude these extensions.
        :param filterLettersExclude: Look for entries that do not contain this string.
        :param filterLettersInclude: Look for entries that contain this string.
        :param filterEndsWith: Look for entries that end with this string. Note that the
            extension should be included with the string. ..i.e. "oom.rad"
        :param filterStartsWith: Look for entries that start with this string. The searching
            for this begins at the file level only and not the directory level.
        :param letterSearchCaseSensitive: If strings are being searched then setting this to
            False makes the search case insensitive.
        :param filterExtensionsCaseSensitive: If set to True then filterLettersInclude and filterLettersExclude options will
            be considered on a case-sensitive basis.
        :param sortByName: If set to True, then sort the filenames by alphabetical order
        :return: List of files/directories in folder dirName.
        """
        files = DirectoryPaths.filesInDirRelative(self.directoryName,
                                                  filesOnly=filesOnly,
                                                  dirsOnly=dirsOnly,
                                                  filterExtensionIncl=filterExtensionIncl,
                                                  filterExtensionExcl=filterExtensionExcl,
                                                  filterLettersExclude=filterLettersExclude,
                                                  filterLettersInclude=filterLettersInclude,
                                                  filterEndsWith=filterEndsWith,
                                                  filterStartsWith=filterStartsWith,
                                                  letterSearchCaseSensitive=letterSearchCaseSensitive,
                                                  sortByName=sortByName,
                                                  filterExtensionsCaseSensitive=filterExtensionsCaseSensitive)
        return files

    def fileDictionary(self, returnAbsPaths=True, filterExtensionIncl=None,
                       filterExtensionExcl=None, filterLettersExclude=None,
                       filterLettersInclude=None, filterEndsWith=None,
                       filterStartsWith=None, letterSearchCaseSensitive=True,filterExtensionsCaseSensitive=False):
        r"""
        Return a dictionary containing keys as fileNames (only) and values as filePaths.
        Absolute paths will be returned by default, though this can be changed.
        For a folder "/home/room/geometry" containing 'floor.rad', 'ceiling.rad', 'walls.rad'
        the output will be {'floor':'/home/room/geometry/floor.rad',
                            'ceiling':'/home/room/geometry/ceiling.rad',
                            'walls':'/home/room/geometry/walls.rad'}

        :param filterExtensionIncl: Include these extensions.
        :param filterExtensionExcl: Exclude these extensions.
        :param filterLettersExclude: Look for entries that do not contain this string.
        :param filterLettersInclude: Look for entries that contain this string.
        :param filterEndsWith: Look for entries that end with this string. Note that the
            extension should be included with the string. ..i.e. "oom.rad"
        :param filterStartsWith: Look for entries that start with this string. The searching
            for this begins at the file level only and not the directory level.
        :param letterSearchCaseSensitive: If strings are being searched then setting this to
            False makes the search case insensitive.
        :param filterExtensionsCaseSensitive: If set to True then filterLettersInclude and filterLettersExclude options will
            be considered on a case-sensitive basis.
        :return: List of files/directories in folder dirName.

        """
        files = DirectoryPaths.fileNamePathDictionary(self.directoryName,
                                                      returnAbsPaths=returnAbsPaths,
                                                      filterExtensionIncl=filterExtensionIncl,
                                                      filterExtensionExcl=filterExtensionExcl,
                                                      filterLettersExclude=filterLettersExclude,
                                                      filterLettersInclude=filterLettersInclude,
                                                      filterEndsWith=filterEndsWith,
                                                      filterStartsWith=filterStartsWith,
                                                      letterSearchCaseSensitive=letterSearchCaseSensitive,
                                                      filterExtensionsCaseSensitive=filterExtensionsCaseSensitive)
        return files

    def filesDuplicate(self, filesOnly=True, dirsOnly=False,
                       filterExtensionIncl=None, filterExtensionExcl=None,
                       filterLettersExclude=None, filterLettersInclude=None,
                       filterEndsWith=None, filterStartsWith=None,
                       letterSearchCaseSensitive=True, returnExtensions=True,filterExtensionsCaseSensitive=False):
        r"""
        Search for duplicates names (only) in a directory. For example if a directory
        room contains [room.rad,room.dwg, room.stl, floor.rad,floor.stl,ground.rad],
        then the output will be {'room':['.stl','.rad','.dwg'],'floor':['.rad','.stl']
        If the returnExtensions option is set to False, then only numbers will be returned.

        :param filesOnly: look for files only.
        :param dirsOnly: look for directories only.
        :param filterExtensionIncl: Include these extensions.
        :param filterExtensionExcl: Exclude these extensions.
        :param filterLettersExclude: Look for entries that do not contain this string.
        :param filterLettersInclude: Look for entries that contain this string.
        :param filterEndsWith: Look for entries that end with this string. Note that the
            extension should be included with the string. ..i.e. "oom.rad"
        :param filterStartsWith: Look for entries that start with this string. The searching
            for this begins at the file level only and not the directory level.
        :param letterSearchCaseSensitive: If strings are being searched then setting this to
            False makes the search case insensitive.
        :param filterExtensionsCaseSensitive: If set to True then filterLettersInclude and filterLettersExclude options will
            be considered on a case-sensitive basis.
        :param returnExtensions: Return extensions of the files containing duplicate names.
        :return: List of files/directories in folder dirName.
        """
        fileDict = DirectoryPaths.findDuplicateNames(self.directoryName,
                                                     filesOnly=filesOnly, dirsOnly=dirsOnly,
                                                     filterExtensionIncl=filterExtensionIncl,
                                                     filterExtensionExcl=filterExtensionExcl,
                                                     filterLettersExclude=filterLettersExclude,
                                                     filterLettersInclude=filterLettersInclude,
                                                     filterEndsWith=filterEndsWith,
                                                     filterStartsWith=filterStartsWith,
                                                     letterSearchCaseSensitive=letterSearchCaseSensitive,
                                                     returnExtensions=returnExtensions,
                                                     filterExtensionsCaseSensitive=filterExtensionsCaseSensitive)
        return fileDict

    def subDirectory(self, name):
        r"""

        :param name: Name of the subdirectory.
        :return: A new class of the type Directory.
        :type Directory
        """
        subDir = os.path.join(self.directoryName, name)
        return Directory(subDir)

    def getSubDirectories(self, filterLettersInclude=None, filterLettersExclude=None,
                          returnDirectoryObjects=True, letterSearchCaseSensitive=True):
        r"""
        Return a list of subdirectories inside this directory.
        :param letterSearchCaseSensitive: Search is case sensitive. Defaults to True.
        :param filterLettersInclude: Return only those subdirectories that match this string. Defaults to None.
        :param filterLettersExclude: Return only those subdirectories that do note match this string. Defaults to None.
        :param returnDirectoryObjects: Return a list of Directory instances. Defaults to True.
        :return:
        """

        assert not (filterLettersExclude and filterLettersInclude), \
            'Both fliterLettersExclude and filterLettersInclude have been set to True. Only one of them can be set to ' \
            'True.'

        putStringInList = lambda x: [x] if isinstance(x, string_types__) else list(x) if x else None

        filterLettersInclude=putStringInList(filterLettersInclude)

        filterLettersExclude=putStringInList(filterLettersExclude)



        dirPath = self.absPath
        subDirs = [path for path in os.listdir(dirPath) if os.path.isdir(os.path.join(dirPath, path))]

        # Filter based on letters.
        # If any of the letters are present, then include in the result.
        if filterLettersInclude:
            if letterSearchCaseSensitive:
                subDirNew = subDirs
            else:
                subDirNew = [val.lower() for val in subDirs]
                filterLettersInclude = [val.lower() for val in filterLettersInclude]

            filteredSubDirs = []
            for idx, subdir in enumerate(subDirNew):
                for filtLet in filterLettersInclude:
                    if filtLet in subdir:
                        filteredSubDirs.append(subDirs[idx])
            subDirs = list(filteredSubDirs)

        #If any of the letters are present, then exclude from the result.
        elif filterLettersExclude:
            if letterSearchCaseSensitive:
                subDirNew = subDirs
            else:
                subDirNew = [val.lower() for val in subDirs]
                filterLettersExclude = [val.lower() for val in filterLettersExclude]

            filteredSubDirs = []
            for idx, subdir in enumerate(subDirNew):
                rejectPath=False
                for filtLet in filterLettersExclude:
                    if filtLet in subdir:
                        rejectPath=True
                if not rejectPath:
                    filteredSubDirs.append(subDirs[idx])
            subDirs = list(filteredSubDirs)



        subDirs = [os.path.join(dirPath, val) for val in subDirs]

        if returnDirectoryObjects:
            subDirs = [Directory(val) for val in subDirs]

        return subDirs

    def replicateDirectoryStructure(self, newDirectoryPath, copyFilesToo=True, directorySizeFilterInMB=None,
                                    overwriteExistingFiles=False,
                                    copyNewOnly=False,
                                    includeRadianceFilesOnly=False, includeExtraRadianceToo=False,
                                    includeRadianceFilesList=['.rad', '.vf', '.pts', '.amb', '.stl', '.xml'],
                                    includeExtraRadianceFilesList=['.hdr', '.tmp', '.vec', '.mtx', '.ill', '.smx',
                                                                   '.wea', '.oct', '.epw', '.pic', '.tiff', '.bmp',
                                                                   '.sky'],
                                    extensionsToInclude=[], extensionsToExclude=[], ):
        r"""
        This method will replicate the directory structure in a different destination with few bells and whistles added.
        At its core all the folder will be copied from one destination  to the next. Additionally all the files can
        be copied by setting the copyFilesToo variable to True. The directorySizeFilter can be used to filter out the
        files in those directories taht are too large (note that in this case none of the files in those large directories
        will be copied). By default, existing files will not be copied..this can be disabled by setting the
        overwriteExistingFiles option to True.

        :param newDirectoryPath: The path to which the files should be copied to. Usually if a directory c:\projects is
            being copied to d: only "d:" is enough, instead of "d:\projects".
        :param copyFilesToo: Set this to True to copy files as well. Defaults to True and will copy all files.
        :param directorySizeFilterInMB: If set to a numberic value, directories whose size in MB is larger than this
            value will not have their files copied.
        :param overwriteExistingFiles: Overwrite existing files. Defaults to False (for safety reasons).
        :type extensionsToInclude: list
        :param extensionsToInclude: If specified only these extensions will be included. Default is an empty list.

        :param extensionsToExclude: If specified these extensions will be excluded. Default is an emply list.
        :param includeRadianceFilesOnly: Boolean. If set to True, only Radiance scene files will be copied.
        :param includeExtraRadianceToo: Boolean. If set to Ture, additional radiance files such as matrices, hdrs etc.
            will also be copied.
        :param includeRadianceFilesList: List of Radiance extensions. Check the param doc for default values.
        :param includeExtraRadianceFilesList: List of Radiance file extensions. Check the param doc for default values.
        :return:
        """

        if extensionsToInclude and isinstance(extensionsToInclude, string_types__):
            extensionsToInclude = [extensionsToInclude]

        if extensionsToExclude and isinstance(extensionsToExclude, string_types__):
            extensionsToExclude = [extensionsToExclude]

        currentDirectory = self.absPath
        currentRootName = os.path.split(currentDirectory)[-1]

        if includeRadianceFilesOnly:
            extensionsToInclude = list(includeRadianceFilesList)

        if includeExtraRadianceToo:
            if extensionsToInclude:
                extensionsToInclude.extend(includeExtraRadianceFilesList)
            else:
                includeExtraRadianceToo = includeExtraRadianceFilesList

        assert not (extensionsToExclude and extensionsToInclude), \
            "The inputs for extensionsToInclude(%s) and extensionsToExclude(%s) cannot be set at the same time.!" % \
            (extensionsToInclude, extensionsToExclude)

        assert not currentDirectory == os.path.abspath(newDirectoryPath), \
            'The path to be written %s and the original path %s are the same.' % (currentDirectory, newDirectoryPath)

        if os.path.exists(newDirectoryPath):
            msg = "The directory %s already exists. So the current directory and its subdirectories will be copied inside" \
                  "this directory." % newDirectoryPath
            warnings.warn(msg)
            newDirectoryPath = os.path.join(newDirectoryPath, currentRootName)
        else:
            mainDir, writeToDir = os.path.split(newDirectoryPath)
            if mainDir:
                assert os.path.exists(
                    mainDir), 'The path %s which is the root directory for the destination directory %,' \
                              ' could not be found.' % (mainDir, newDirectoryPath)

                if not os.path.exists(newDirectoryPath):
                    os.mkdir(newDirectoryPath)
                else:
                    warnings.warn("The path %s already exists" % newDirectoryPath)

        existingDirectories = []
        existingFiles = []
        excludedDirectoriesForSize = []
        overWrittenFiles = []
        overWriteWarnings = []
        oldFiles = []
        for dirs, subs, fileNames in os.walk(currentDirectory):
            relativePath = os.path.relpath(dirs, currentDirectory)
            newAbsPath = os.path.join(newDirectoryPath, relativePath)

            if not os.path.exists(newAbsPath):
                os.mkdir(newAbsPath)
            else:
                existingDirectories.append(newAbsPath)

            if copyFilesToo:
                if directorySizeFilterInMB and (not (extensionsToExclude or extensionsToInclude)):
                    if Directory(dirs).sizeMBytesExcludeSub > directorySizeFilterInMB:
                        excludedDirectoriesForSize.append(
                            "%s: %sMB" % (newAbsPath, Directory(dirs).sizeMBytesExcludeSub))
                        continue

                for fileName in fileNames:
                    fileExt = os.path.splitext(fileName)[-1].lower()
                    copyThisFile = False

                    if extensionsToInclude:
                        if fileExt in map(string.lower, extensionsToInclude):
                            copyThisFile = True
                        else:
                            copyThisFile = False
                    elif extensionsToExclude:
                        if fileExt in map(string.lower, extensionsToExclude):
                            copyThisFile = False
                        else:
                            copyThisFile = True
                    elif not (extensionsToInclude or extensionsToExclude):
                        copyThisFile = True

                    if copyThisFile:
                        currName = os.path.join(dirs, fileName)
                        newName = os.path.join(newAbsPath, fileName)
                        if os.path.exists(newName):

                            if copyNewOnly:
                                modTime = os.path.getmtime(newName)
                                modOrigTime = os.path.getmtime(currName)
                                if modTime > modOrigTime:
                                    oldFiles.append("%50s | Src: %s | Dest: %s" % (newName,
                                                                                   time.strftime('%H:%M:%S %m/%d/%Y',
                                                                                                 time.gmtime(
                                                                                                     modOrigTime)),
                                                                                   time.strftime('%H:%M:%S %m/%d/%Y',
                                                                                                 time.gmtime(modTime))))
                                    continue

                            if not overwriteExistingFiles:
                                overWriteWarnings.append(currName)
                            else:
                                shutil.copy(currName, newName)
                                overWrittenFiles.append(newName)
                        else:
                            fileSizeInMB = FileObj(currName).sizeMBytes
                            if fileSizeInMB > 500:
                                msg = "The file %s is larger than %sMB. Copying might take time." % (
                                currName, fileSizeInMB)
                                warnings.warn(msg)
                            shutil.copy(currName, newName)

        if existingDirectories:
            msg = "The following directories already exist and were not created:\n\t%s" % "\n\t".join(
                existingDirectories)
            warnings.warn(msg)
        if excludedDirectoriesForSize:
            msg = "The files in the following directories were not copied because the " \
                  "directory size was larger than %s MB:\n\t%s" % (
                  directorySizeFilterInMB, "\n\t".join(excludedDirectoriesForSize))
            warnings.warn(msg)

        if overWrittenFiles:
            msg = "The following files were overwritten:\n\t%s" % "\n\t".join(overWrittenFiles)
            warnings.warn(msg)

        if overWriteWarnings:
            msg = "The following files exist and were not overwritten (These can be overwritten by setting the " \
                  "variable 'overwriteExistingFiles' to True) :\n\t%s" % "\n\t".join(overWriteWarnings)
            warnings.warn(msg)

        if oldFiles:
            msg = "The following files are apparently newer than the originals and therefore were not copied:\n\t%s" % "\n\t".join(
                oldFiles)
            warnings.warn(msg)

        return True

    def fileInside(self, name):
        """Return the absolute path of the first file that matches the 'name' string"""
        fileName = self.fileNamesAbsolute(filterLettersInclude=name, letterSearchCaseSensitive=True)[0]
        return fileName

    def fileInsideExact(self, name):
        """Return the absolute path of the file that eaxctly matches the 'name' string"""
        fileName = self.fileNamesAbsolute(filterLettersInclude=name, letterSearchCaseSensitive=True)
        fileName = [val for val in fileName if os.path.split(val)[-1] == name]

        assert len(fileName), 'The file %s was not found inside the directory %s' % (name, self.absPath)

        return fileName[0]

    def fileObjInsideExact(self, name):
        """Return the FileObj instance  of the file that eaxctly matches the 'name' string"""
        fileName = self.fileNamesAbsolute(filterLettersInclude=name, letterSearchCaseSensitive=True)
        fileName = [val for val in fileName if os.path.split(val)[-1] == name]

        assert len(fileName), 'The file %s was not found inside the directory %s' % (name, self.absPath)

        return FileObj(fileName[0])

    def fileObjInside(self, name):
        """Return a FileObj instance of the first file that matches the 'name' string"""
        fileName = self.fileNamesAbsolute(filterLettersInclude=name, letterSearchCaseSensitive=True)[0]
        return FileObj(fileName)

    @property
    def _calculateSize(self):
        """Calculate the size of this directory in Bytes."""
        currPath = self.absPath
        total_size = 0
        for dirpath, dirnames, filenames in os.walk(currPath):
            for f in filenames:
                try:
                    fp = os.path.join(dirpath, f)
                    total_size += os.path.getsize(fp)
                except WindowsError:
                    logger.warn("The file %s was ignored due to incorrect syntax." % f)
        return total_size

    @property
    def _calculateSizeExcludeSubDir(self):
        """Calculate the size of only the filese in this directory in Bytes."""
        currPath = self.absPath
        total_size = 0
        for filenames in os.listdir(currPath):
            fp = os.path.join(currPath, filenames)
            if os.path.isfile(fp):
                total_size += os.path.getsize(fp)
        return total_size

    @property
    def sizeBytes(self):
        """Size in bytes"""
        return self._calculateSize

    @property
    def sizeKBytes(self):
        """Size in kilobytes rounded off to single decimal place."""
        return round(self._calculateSize / 1024, 1)

    @property
    def sizeMBytes(self):
        """Size in megabytes rounded off to four decimal places."""

        return round(self._calculateSize / (1024 * 1024), 4)

    @property
    def sizeGBytes(self):
        """Size in GB rounded off to four decimal places."""
        return round(self._calculateSize / (1024 * 1024 * 1024), 4)

    @property
    def sizeBytesExcludeSub(self):
        """Size in bytes but exclude subdirectories"""
        return self._calculateSizeExcludeSubDir

    @property
    def sizeKBytesExcludeSub(self):
        """Size in kilobytes rounded off to single decimal place but exclude subdirectories"""
        return round(self._calculateSizeExcludeSubDir / 1024, 1)

    @property
    def sizeMBytesExcludeSub(self):
        """Size in megabytes rounded off to four decimal places but exclude subdirectories"""

        return round(self._calculateSizeExcludeSubDir / (1024 * 1024), 4)

    @property
    def sizeGBytesExcludeSub(self):
        """Size in GB rounded off to four decimal places but exclude subdirectories"""
        return round(self._calculateSizeExcludeSubDir / (1024 * 1024 * 1024), 4)

    @property
    def absPath(self):
        return os.path.abspath(self.directoryName)

    @property
    def relPath(self):
        return os.path.relpath(self.directoryName)

    def __str__(self):
        return str(self.directoryName)

    def __repr__(self):
        return str(self.directoryName)

    def __unicode__(self):
        return self.directoryName

    @property
    def path(self):
        return str(self.directoryName)

    @property
    def timeModified(self):
        r"""
        Return the last modified time in datetime.datetime format.
        @:rtype: datetime.datetime

        :return:
        """
        return datetime.datetime.fromtimestamp(os.path.getmtime(self.absPath))

    @property
    def timeCreated(self):
        r"""
        Return the last modified time in datetime.datetime format.

        Try to get the date that a file was created, falling back to when it was
        last modified if that isn't possible.
        See http://stackoverflow.com/a/39501288/1709587 for explanation.

        @:rtype: datetime.datetime
        """
        if platform.system() == 'Windows':
            return datetime.datetime.fromtimestamp(os.path.getctime(self.absPath))
        else:
            stat = datetime.datetime.fromtimestamp(os.stat(self.absPath))
            try:
                return datetime.datetime.fromtimestamp(stat.st_birthtime)
            except AttributeError:
                # We're probably on Linux. No easy way to get creation dates here,
                # so we'll settle for when its content was last modified.
                return datetime.datetime.fromtimestamp(stat.st_mtime)

    def searchWord(self, words, includeLineInResults=False, matchCase=True, matchExact=False,
                   extInclude=[], extExclude=[], fileNameStrInclude=[], fileNameStrExclude=[],
                   fileNameMatchExact=False, dirNameStrInclude=[], dirNameStrExclude=[],
                   dirNameMatchExact=False, useExcludeDefaultExts=True, useIncludeDefaultExts=False,
                   warnErrors=True, binaryCheck=True, printIgnoredBinaryFiles=False):
        r"""
        Search for a single word or multiple words in this directory. Default
        option is to return the line number on which the word occurs. If the includeLineInResults
        is set to True, then the line will be included as well.

        :param sources: This can be a list of files or directories. If directories are provided,
            then the script will search the files inside those directories recursively.
        :param words: A single word or a list of words.
        :param includeLineInResults: Include the line on which the word's occur in the results.
            Default is False.
        :param matchCase: Case sensitive.
        :param matchExact: Word(s) should be exact matches.
        :param extInclude: File extensions to include.
        :param extExclude: File extensions to exclude.
        :param fileNameStrInclude: Include file names containing these strings.
        :param fileNameStrExclude: Exclude file names containing these string.
        :param fileNameMatchExact: The fileName should be exact.
        :param dirNameStrInclude: Include directory names containing these strings.
        :param dirNameStrExclude: Exclude directory names containing these strings.
        :param dirNameMatchExact: The directory name match should be exact.
        :param useExcludeDefaultExts: Some default files such as .exe, .png etc have been
            already excluded. Set this to False to consider them as well.
        :param useIncludeDefaultExts:Some default files such as .txt, .log etc have been
            already included. Set this to False to ignore those defaults.
        :param warnErrors: Raising warnings in cases of errors.
        :return
        """

        return _searchWord(self.directoryName, words, includeLineInResults=includeLineInResults,
                           matchCase=matchCase, matchExact=matchExact, extInclude=extInclude,
                           extExclude=extExclude, fileNameStrInclude=fileNameStrInclude,
                           fileNameStrExclude=fileNameStrExclude, fileNameMatchExact=fileNameMatchExact,
                           dirNameStrInclude=dirNameStrInclude, dirNameStrExclude=dirNameStrExclude,
                           dirNameMatchExact=dirNameMatchExact, useExcludeDefaultExts=useExcludeDefaultExts,
                           useIncludeDefaultExts=useIncludeDefaultExts, warnErrors=warnErrors,
                           binaryCheck=binaryCheck, printIgnoredBinaryFiles=printIgnoredBinaryFiles)

    def filterFilesByTimeModified(self, afterTime=None, beforeTime=None,
                                  filterExtensionIncl=None, filterExtensionExcl=None,
                                  filterLettersExclude=None, filterLettersInclude=None,
                                  filterEndsWith=None, filterStartsWith=None,
                                  letterSearchCaseSensitive=True, returnRelativePaths=False,
                                 sortByName=True,filterExtensionsCaseSensitive=False):

        """
        Filter the files in a directory based on creation time.

        :param afterTime: Filter files that were created after this time. The input for this can either be an
             instance of datetime or simply a string (like 'yesterday')
        :param beforeTime: Filter files that were created before this time. The input for this can either be an
             instance of datetime or simply a string (like 'tomorrow', "three days ago" etc.)
        :param filterExtensionIncl: Include these extensions.
        :param filterExtensionExcl: Exclude these extensions.
        :param filterLettersExclude: Look for entries that do not contain this string.
        :param filterLettersInclude: Look for entries that contain this string.
        :param filterEndsWith: Look for entries that end with this string. Note that the
            extension should be included with the string. ..i.e. "oom.rad"
        :param filterStartsWith: Look for entries that start with this string. The searching
            for this begins at the file level only and not the directory level.
        :param letterSearchCaseSensitive: If strings are being searched then setting this to
            False makes the search case insensitive.
        :param filterExtensionsCaseSensitive: If set to True then filterLettersInclude and filterLettersExclude options will
            be considered on a case-sensitive basis.
        :param sortByName: If set to True, then sort the filenames by alphabetical order
        :return:
        """

        allFiles = self.fileNamesAbsolute(filesOnly=True, filterExtensionExcl=filterExtensionExcl,
                                          filterExtensionIncl=filterExtensionIncl,
                                          filterLettersExclude=filterLettersExclude,
                                          filterLettersInclude=filterLettersInclude,
                                          filterEndsWith=filterEndsWith, filterStartsWith=filterStartsWith,
                                          letterSearchCaseSensitive=letterSearchCaseSensitive,sortByName=sortByName,
                                          filterExtensionsCaseSensitive=filterExtensionsCaseSensitive)
        allObjs = map(FileObj, allFiles)

        if afterTime:
            try:
                afterTime = Calendar().parseReturnDateTime(afterTime)
            except:
                assert isinstance(afterTime, datetime.datetime), \
                    "The input for after time should either be a string that can be converted into a datetime instance" \
                    " or it should be a datetime instance."
        if beforeTime:
            try:
                beforeTime = Calendar().parseReturnDateTime(beforeTime)
            except:
                assert isinstance(beforeTime, datetime.datetime), \
                    "The input for before time should either be a string that can be converted into a datetime instance" \
                    " or it should be a datetime instance."

        # if before and affter are provided, ensure that it is an interval!
        if beforeTime and afterTime:
            if beforeTime < afterTime:
                beforeTime, afterTime = afterTime, beforeTime

        fileList = []
        for fileObj in allObjs:
            if beforeTime and not afterTime:
                if fileObj.timeModified < beforeTime:
                    fileList.append(fileObj)

            if afterTime and not beforeTime:
                if fileObj.timeModified > afterTime:
                    fileList.append(fileObj)

            if beforeTime and afterTime:
                if afterTime < fileObj.timeModified < beforeTime:
                    fileList.append(fileObj)

            if not (beforeTime or afterTime):
                fileList.append(fileObj)

        # Rel path does not always work!
        if fileList and returnRelativePaths:
            fileList = [fileObj.relPath for fileObj in fileList]
        else:
            fileList = [fileObj.absPath for fileObj in fileList]

        return fileList

    def filterFilesByTimeCreated(self, afterTime=None, beforeTime=None,
                                 filterExtensionIncl=None, filterExtensionExcl=None,
                                 filterLettersExclude=None, filterLettersInclude=None,
                                 filterEndsWith=None, filterStartsWith=None,
                                 letterSearchCaseSensitive=True, returnRelativePaths=False,
                                 sortByName=True,filterExtensionsCaseSensitive=False):

        """
        Filter the files in a directory based on creation time.

        :param afterTime: Filter files that were created after this time. The input for this can either be an
             instance of datetime or simply a string (like 'yesterday')
        :param beforeTime: Filter files that were created before this time. The input for this can either be an
             instance of datetime or simply a string (like 'tomorrow', "three days ago" etc.)
        :param filterExtensionIncl: Include these extensions.
        :param filterExtensionExcl: Exclude these extensions.
        :param filterLettersExclude: Look for entries that do not contain this string.
        :param filterLettersInclude: Look for entries that contain this string.
        :param filterEndsWith: Look for entries that end with this string. Note that the
            extension should be included with the string. ..i.e. "oom.rad"
        :param filterStartsWith: Look for entries that start with this string. The searching
            for this begins at the file level only and not the directory level.
        :param letterSearchCaseSensitive: If strings are being searched then setting this to
            False makes the search case insensitive.
        :param filterExtensionsCaseSensitive: If set to True then filterLettersInclude and filterLettersExclude options will
            be considered on a case-sensitive basis.
        :param sortByName: If set to True, then sort the filenames by alphabetical order
        :return:
        """

        allFiles = self.fileNamesAbsolute(filesOnly=True, filterExtensionExcl=filterExtensionExcl,
                                          filterExtensionIncl=filterExtensionIncl,
                                          filterLettersExclude=filterLettersExclude,
                                          filterLettersInclude=filterLettersInclude,
                                          filterEndsWith=filterEndsWith, filterStartsWith=filterStartsWith,
                                          letterSearchCaseSensitive=letterSearchCaseSensitive,sortByName=sortByName,
                                          filterExtensionsCaseSensitive=filterExtensionsCaseSensitive)
        allObjs = map(FileObj, allFiles)

        if afterTime:
            try:
                afterTime = Calendar().parseReturnDateTime(afterTime)
            except:
                assert isinstance(afterTime, datetime.datetime), \
                    "The input for after time should either be a string that can be converted into a datetime instance" \
                    " or it should be a datetime instance."
        if beforeTime:
            try:
                beforeTime = Calendar().parseReturnDateTime(beforeTime)
            except:
                assert isinstance(beforeTime, datetime.datetime), \
                    "The input for before time should either be a string that can be converted into a datetime instance" \
                    " or it should be a datetime instance."

        # if before and affter are provided, ensure that it is an interval!
        if beforeTime and afterTime:
            if beforeTime < afterTime:
                beforeTime, afterTime = afterTime, beforeTime

        fileList = []
        for fileObj in allObjs:
            if beforeTime and not afterTime:
                if fileObj.timeCreated < beforeTime:
                    fileList.append(fileObj)

            if afterTime and not beforeTime:
                if fileObj.timeCreated > afterTime:
                    fileList.append(fileObj)

            if beforeTime and afterTime:
                if afterTime < fileObj.timeCreated < beforeTime:
                    fileList.append(fileObj)

            if not (beforeTime or afterTime):
                fileList.append(fileObj)

        # Rel path does not always work!
        if fileList and returnRelativePaths:
            fileList = [fileObj.relPath for fileObj in fileList]
        else:
            fileList = [fileObj.absPath for fileObj in fileList]

        return fileList

    def filterFilesBySize(self,sizeGreaterThanEqual=None,sizeLessThanEqual=None,
                                sizeMeasuredIn="KB", filterExtensionIncl=None, filterExtensionExcl=None,
                                 filterLettersExclude=None, filterLettersInclude=None,
                                 filterEndsWith=None, filterStartsWith=None,
                                 letterSearchCaseSensitive=True, returnRelativePaths=False,
                                 sortByName=True,filterExtensionsCaseSensitive=False):
        """
        Filter file sizes that are greater than or equal to a certain value. The default value for measuring is kilobytes,
        however, this can also be bytes, MB and GB.
        :param sizeGreaterThanEqual: Include only those files that are greater than or equal to this value.
        :param sizeLessThanEqual: Include only those files that are less than or equal to this value.
        :param sizeMeasuredIn: Defaults to KB or Kilobytes. Can also be 'b' bytes, 'MB' megabytes, 'GB' gigabytes. The
            letters can be upper or lower case
        :param filterExtensionIncl: Include these extensions.
        :param filterExtensionExcl: Exclude these extensions.
        :param filterLettersExclude: Look for entries that do not contain this string.
        :param filterLettersInclude: Look for entries that contain this string.
        :param filterEndsWith: Look for entries that end with this string. Note that the
            extension should be included with the string. ..i.e. "oom.rad"
        :param filterStartsWith: Look for entries that start with this string. The searching
            for this begins at the file level only and not the directory level.
        :param letterSearchCaseSensitive: If strings are being searched then setting this to
            False makes the search case insensitive.
        :param filterExtensionsCaseSensitive: If set to True then filterLettersInclude and filterLettersExclude options will
            be considered on a case-sensitive basis.
        :param sortByName: If set to True, then sort the filenames by alphabetical order.
        :return:
        """

        assert (sizeLessThanEqual is not None) or (sizeGreaterThanEqual is not None),\
            "For the filtering to work both 'sizeLessThanEqual' and 'sizeGreaterThanEqual' cannot be None."

        if (sizeLessThanEqual is not None):
            assert sizeLessThanEqual>=0,"The value for the parameter sizeLessThanEqual(%s) cannot be less than 0"%sizeLessThanEqual

        if (sizeGreaterThanEqual is not None):
            assert sizeGreaterThanEqual>=0,"The value for the parameter sizeGreaterThanEqual(%s) cannot be less than 0"%sizeGreaterThanEqual

        #If no greater than value is given then start counting at zero and greater. This will ensure that all are included.
        sizeGreaterThanEqual=sizeGreaterThanEqual if sizeGreaterThanEqual is not None else 0

        #If no less than value is given then start counting at 10**15 bytes, roughly 1000TB I think.
        sizeLessThanEqual=sizeLessThanEqual if sizeLessThanEqual is not None else 10**15

        if sizeGreaterThanEqual>sizeLessThanEqual:
            logger.warning("Flipped the values for 'sizeGreaterThanEqual'(%s) and 'sizeLessThanEqual'(%s)."%(sizeGreaterThanEqual,sizeLessThanEqual))
            sizeGreaterThanEqual,sizeLessThanEqual=sizeLessThanEqual,sizeGreaterThanEqual

        sizeMeasuredIn=sizeMeasuredIn.lower()

        measureDict={"b":(1,"bytes"),"kb":(1024,"kiloBytes"),"mb":(1024**2,"megaBytes"),"gb":(1024**3,"gigaBytes")}


        assert sizeMeasuredIn in measureDict.keys(),"The value of the parameter 'sizeMeasuredIn' must be one of the following:%s"%",".join(measureDict.keys())

        sizeDivider=measureDict[sizeMeasuredIn][0]

        allFiles = self.fileNamesAbsolute(filesOnly=True, filterExtensionExcl=filterExtensionExcl,
                                          filterExtensionIncl=filterExtensionIncl,
                                          filterLettersExclude=filterLettersExclude,
                                          filterLettersInclude=filterLettersInclude,
                                          filterEndsWith=filterEndsWith, filterStartsWith=filterStartsWith,
                                          letterSearchCaseSensitive=letterSearchCaseSensitive,sortByName=sortByName,
                                          filterExtensionsCaseSensitive=filterExtensionsCaseSensitive)
        allObjs = list(map(FileObj, allFiles))

        fileList=[]

        for fobj in allObjs:
            if sizeGreaterThanEqual<=fobj.sizeBytes/sizeDivider<=sizeLessThanEqual:
                fileList.append(fobj)

        logger.debug("Filtered %s files out of %s based on size limits of %s%s and %s%s."%(len(fileList),len(allObjs),sizeGreaterThanEqual,sizeMeasuredIn,
                                                                                           sizeLessThanEqual,sizeMeasuredIn))

        # Rel path does not always work!
        if fileList and returnRelativePaths:
            fileList = [fileObj.relPath for fileObj in fileList]
        else:
            fileList = [fileObj.absPath for fileObj in fileList]

        return fileList


    def deleteFiles(self, modifiedAfter=None, modifiedBefore=None, createdAfter=None, createdBefore=None,
                    filterExtensionIncl=None, filterExtensionExcl=None,
                    filterLettersExclude=None, filterLettersInclude=None,
                    filterEndsWith=None, filterStartsWith=None,
                    letterSearchCaseSensitive=True):

        assert not ((modifiedAfter or modifiedBefore) and (createdAfter or createdBefore)), \
            "Can only delete by filtering either according to modified times or creation times. Not both!"

        if modifiedAfter or modifiedBefore:
            fileList = self.filterFilesByTimeModified(afterTime=modifiedAfter, beforeTime=modifiedBefore,
                                                      filterExtensionIncl=filterExtensionIncl,
                                                      filterExtensionExcl=filterExtensionExcl,
                                                      filterLettersExclude=filterLettersExclude,
                                                      filterLettersInclude=filterLettersInclude,
                                                      filterEndsWith=filterEndsWith, filterStartsWith=filterStartsWith,
                                                      letterSearchCaseSensitive=letterSearchCaseSensitive)
        else:
            fileList = self.filterFilesByTimeCreated(afterTime=createdAfter, beforeTime=createdBefore,
                                                     filterExtensionIncl=filterExtensionIncl,
                                                     filterExtensionExcl=filterExtensionExcl,
                                                     filterLettersExclude=filterLettersExclude,
                                                     filterLettersInclude=filterLettersInclude,
                                                     filterEndsWith=filterEndsWith,
                                                     filterStartsWith=filterStartsWith,
                                                     letterSearchCaseSensitive=letterSearchCaseSensitive)

        for fileName in fileList:
            send2trash(fileName)


    def fileNamesAbsolutePlus(self,createdBefore=None,createdAfter=None,modifiedAfter=None,
                                modifiedBefore=None,sizeGreaterThanOrEqual=None,sizeLessThanOrEqual=None,sizeMeasuredIn="KB",
                                filterExtensionIncl=None, filterExtensionExcl=None,
                                filterLettersExclude=None, filterLettersInclude=None,
                                filterEndsWith=None, filterStartsWith=None,
                                letterSearchCaseSensitive=True, sortByName=True, filterExtensionsCaseSensitive=False):
        """
        Given a directory, return all the files on the root level based on extensive filtering criteria. This is a much
        souped up version of the favorite function fileNamesAbsolute. If filtering is done according to modified dates,
        created dates and/or size, then the order of filtering is created dates, modified dates and size.

        :param modifiedAfter: Filter files that were created after this time. The input for this can either be an
             instance of datetime or simply a string (like 'yesterday')
        :param modifiedBefore: Filter files that were created before this time. The input for this can either be an
             instance of datetime or simply a string (like 'tomorrow', "three days ago" etc.)
        :param createdAfter: Filter files that were created after this time. The input for this can either be an
            instance of datetime or simply a string (like 'yesterday')
        :param createdBefore: Filter files that were created before this time. The input for this can either be an
            instance of datetime or simply a string (like 'tomorrow', "three days ago" etc.)
        :param sizeGreaterThanOrEqual: Include only those files that are greater than or equal to this value.
        :param sizeLessThanOrEqual: Include only those files that are less than or equal to this value.
        :param sizeMeasuredIn: Defaults to KB or Kilobytes. Can also be 'b' bytes, 'MB' megabytes, 'GB' gigabytes. The
         letters can be upper or lower case
        :param filterExtensionIncl: Include these extensions.
        :param filterExtensionExcl: Exclude these extensions.
        :param filterLettersExclude: Look for entries that do not contain this string.
        :param filterLettersInclude: Look for entries that contain this string.
        :param filterEndsWith: Look for entries that end with this string. Note that the
            extension should be included with the string. ..i.e. "oom.rad"
        :param filterStartsWith: Look for entries that start with this string. The searching
            for this begins at the file level only and not the directory level.
        :param letterSearchCaseSensitive: If strings are being searched then setting this to
            False makes the search case insensitive.
        :param filterExtensionsCaseSensitive: If set to True then filterLettersInclude and filterLettersExclude options will
            be considered on a case-sensitive basis.
        :param sortByName: If set to True, then sort the filenames by alphabetical order

        :return: A list of files that matches the search criteria.

        """
        kwargs = {"filterExtensionIncl": filterExtensionIncl,
                              "filterExtensionExcl": filterExtensionExcl,
                              "filterLettersExclude": filterLettersExclude,
                              "filterLettersInclude": filterLettersInclude,
                              "filterEndsWith": filterEndsWith,
                              "filterStartsWith": filterStartsWith,
                              "letterSearchCaseSensitive": letterSearchCaseSensitive,
                              "sortByName": sortByName,
                              "filterExtensionsCaseSensitive": filterExtensionsCaseSensitive}

        logger.debug("Checking files for the directory %s"%self.absPath)

        #We need this one because it is possible that the other filtering options are not considered.
        fileList=self.fileNamesAbsolute(**kwargs)

        logger.debug("Filtered %s files based on name and extension alone."%len(fileList))

        logMsg="Further filtered"

        if createdBefore or createdAfter:
            fileList=self.filterFilesByTimeCreated(afterTime=createdAfter,beforeTime=createdBefore,**kwargs)

            logMsg+=" %s files based on creation date"%len(fileList)

        if modifiedBefore or modifiedAfter:
            fileListModified=self.filterFilesByTimeModified(afterTime=modifiedAfter,beforeTime=modifiedBefore,**kwargs)


            #Keep only those files in this list which are present in fileListCreated if it was checked
            # fileListCreated was initialized as None, if it was considered in filtering then
            # it will become a list, empty or otherwise.
            if createdBefore or createdAfter:
                fileList=[filePath for filePath in fileListModified if filePath in fileList]
            else:
                fileList=list(fileListModified)

            logMsg += ", %s files based on modified date" % len(fileList)

        if sizeLessThanOrEqual or sizeGreaterThanOrEqual:
            fileListSize=self.filterFilesBySize(sizeLessThanEqual=sizeLessThanOrEqual,sizeGreaterThanEqual=sizeGreaterThanOrEqual,sizeMeasuredIn=sizeMeasuredIn,
                                                              **kwargs)


            if modifiedAfter or modifiedBefore or createdBefore or createdAfter:
                fileList=[filePath for filePath in fileListSize if filePath in fileList]
            else:
                fileList=list(fileListSize)
            logMsg += ", %s files based on size" % len(fileList)

        if modifiedAfter or modifiedBefore or createdBefore or createdAfter or sizeGreaterThanOrEqual or sizeLessThanOrEqual:
            logger.debug(logMsg)

        return fileList



    def walkRecurFilter(self, depth = 0, filterSubDirLettersInclude=None, filterSubDirLettersExclude=None,
                    subDirletterSearchCaseSensitive=True, subDirfilterAppliesToEntirePath=False,retRootAsDirObj=False,
                        retSubListAsDirObjList=False,retFileListAsFileObjList=False):
        r"""
        An implementation of os.walk wherein the recursion depth, ie. how deep the folder level should be 'walked' can
        be specified with the depth parameter. Additionally, the directories being returned can be filtered
        based on letters.

        Usage:
            >>> someDir= Directory("c:\somedir")
            >>> for roots, subs, fileLists in someDir.walkRecurFilter(depth=2):
            # This can then be used as a standard os.walk function.

        :param depth: The recursion depth that is to be followed. 0 will return the current directory, any value above 0
            will recurse to that value. -1 will go through the entire directory.
        :param filterSubDirLettersExclude: Only return those directories which do not include these letters. Defaults
            to None, but can be a single string or list of strings.
        :param filterSubDirLettersExclude: Only return those directories which include these letters. Defaults
            to None, but can be a single string or list of strings.
        :param subDirletterSearchCaseSensitive: If set to True, then the directory string matching for inclusion and
            exclusion testing will be case sensitive.
        :param subDirfilterAppliesToEntirePath: If set to True, then the inclusion and exclusion testing will be
            applied to the entire path of the directory and not just the last element of the path. i.e. If true
            then all of 'c:\users\ladybug\funstuff' will be considered, but if set to False, then only 'funstuff' will be
            considered.
        :param retFileListAsFileObjList: If set to true, the fileLists in root,subs,fileLists will be returned as a
            list of FileObjs instead of a list of strings.
        :param retSubListAsDirObjList: If set to true, the subs in root,subs,fileLists will be returned as a list
            of Directory instances instead of a list of strings.
        :param retRootAsDirObj: If set to true, the root in root,subs,fileLists will be returned as a Directory object.
        :return: A generator object will make it possible to iterate through files and folders the same way as os.walk.
        """

        path=self.absPath

        putStringInList = lambda x: [x] if isinstance(x, str) else list(
            x) if x else None

        #Depending on the test condition, either return the root path as string or as a an instance of Directory.
        dirClsOrStr = lambda pth, boolTest: Directory(pth) if boolTest else pth

        #Depending on the test condition, either return the subDir paths as the original list or as a list of Directory instances.
        subListDirClsOrStrList=lambda subList,boolTest,rootPth: list(map(Directory,[os.path.join(rootPth,val) for val in subList])) if boolTest else subList

        #Depending on the test condition, either return the fileList paths as the original list or as a list of FileObj instances.
        fileListFileObjClsOrStrList=lambda fileList,boolTest,rootPth: list(map(FileObj,[os.path.join(rootPth,val) for val in fileList])) if boolTest else fileList


        filterSubDirLettersInclude = putStringInList(filterSubDirLettersInclude)

        filterSubDirLettersExclude = putStringInList(filterSubDirLettersExclude)


        rootDirLogMsg="root directory as an instance of Directory" if retRootAsDirObj else "root directory as a string (default)"
        subDirLogMsg="sub directory list as a list of instances of Directory" if retSubListAsDirObjList else "sub directory list as a list of string(default)"
        fileLogMsg="file list as a list of instances of FileObj" if retFileListAsFileObjList else "file list as a list of strings (default)"

        logger.debug("Will be returning %s, %s and %s."%(rootDirLogMsg,subDirLogMsg,fileLogMsg))

        logMsgLetter="Letter search is %s"%("case sensitive" if subDirletterSearchCaseSensitive else "not case sensitive.")

        if not subDirletterSearchCaseSensitive:
            if filterSubDirLettersExclude:
                filterSubDirLettersExclude = [val.lower() for val in filterSubDirLettersExclude]
            if filterSubDirLettersInclude:
                filterSubDirLettersInclude = [val.lower() for val in filterSubDirLettersInclude]


        if filterSubDirLettersInclude:
            logMsgLetter+=" The following text strings will be considered for inclusion test: %s"%",".join(filterSubDirLettersInclude)

        if filterSubDirLettersExclude:
            logMsgLetter+=" The following text strings will be considered for exclusion test: %s"%",".join(filterSubDirLettersExclude)

        logger.debug(logMsgLetter)
        # If depth is negative, just walk
        # Not using yield from for python2 compat
        # and copy dirs to keep consistant behavior for depth = -1 and depth = inf
        if depth < 0:
            # Note: I changed this to 30,000 or rather some really high number
            depth = 30000
            # for root, dirs, files in os.walk(path):
            #     yield root, dirs[:], files
            # return
        #elif depth == 0:
         #   return self.absPath

        # path.count(os.path.sep) is safe because
        # - On Windows "\\" is never allowed in the name of a file or directory
        # - On UNIX "/" is never allowed in the name of a file or directory
        # - On MacOS a literal "/" is quitely translated to a ":" so it is still
        #   safe to count "/".
        base_depth = path.rstrip(os.path.sep).count(os.path.sep)
        for root, dirs, files in os.walk(path):
            # This is to check
            if subDirfilterAppliesToEntirePath:
                folderStrToCheck = root
            else:
                parent, folderStrToCheck = os.path.split(root)
                folderStrToCheck = folderStrToCheck if folderStrToCheck else parent

            folderStrToCheck = folderStrToCheck if subDirletterSearchCaseSensitive else folderStrToCheck.lower()

            if filterSubDirLettersInclude:
                pathYielded=False
                for filtLet in filterSubDirLettersInclude:
                    if filtLet in folderStrToCheck and not pathYielded:
                        pathYielded=True
                        yield dirClsOrStr(root, retRootAsDirObj), subListDirClsOrStrList(dirs[:],retSubListAsDirObjList,root), fileListFileObjClsOrStrList(files,retFileListAsFileObjList,root)
            elif filterSubDirLettersExclude:
                exclFound = False
                for filtLet in filterSubDirLettersExclude:
                    if filtLet in folderStrToCheck:
                        exclFound = True
                        break
                if not exclFound:
                    yield dirClsOrStr(root, retRootAsDirObj), subListDirClsOrStrList(dirs[:],retSubListAsDirObjList,root), fileListFileObjClsOrStrList(files,retFileListAsFileObjList,root)
            else:
                yield dirClsOrStr(root, retRootAsDirObj), subListDirClsOrStrList(dirs[:],retSubListAsDirObjList,root), fileListFileObjClsOrStrList(files,retFileListAsFileObjList,root)

            cur_depth = root.count(os.path.sep)
            if base_depth + depth <= cur_depth:
                del dirs[:]


class RootDirectory(Directory):
    r"""
    This class is useful to assign paths as per a specified as specified root directory.
    The _init_ requires a directory path

    Useful methods:
        changeToRoot: Will set the current working directory to rootDirectory while
            remembering the cwd at the time.
        changeBackFromRoot: Will set the cwd to whatever it was when it was changed from
            root.

    This class inherits from FilePaths which contains a lot of extremely useful
    classmethods. Read the documentation for that.

    """

    def __init__(self, filePath, rootDirectory,directoryVariableName=None):
        r"""
        A useful class for generating paths relating to the current file.
        :param rootDirectory: This is directory that is the root path for the project.
        :param filePath: A file that exists in the root directory.
        :param directoryVariableName: A variable name of for the directory, that
         can be used for assertion testing.
        """
        self.rootDirectory = rootDirectory
        self.filePath = filePath
        self._currentDirectory = os.path.dirname(filePath)

        Directory.__init__(self, self.rootDirectory,directoryVariableName=directoryVariableName)

    @property
    def filePath(self):
        return self._filePath

    @filePath.setter
    def filePath(self, value):
        self._filePath = os.path.abspath(value)

    def changeToRoot(self):
        """Change the current working directory to the rootDirectory"""
        self._currentDirectory = os.getcwd()
        os.chdir(self.rootDirectory)

    def changeBackFromRoot(self):
        r"""Change the current workind directory from root to whatever it was
        before."""
        os.chdir(self._currentDirectory)

    @property
    def rootDirectory(self):
        return self._rootDirectory

    @rootDirectory.setter
    def rootDirectory(self, rootPath):
        r"""The path can be an absolute path (in which it will be returned) or it can be
         relative. If it is a single word then the script search from the current cwd till
         the directory ends with that word."""
        if os.path.isabs(rootPath):
            self._rootDirectory = rootPath
        else:
            currDir = os.getcwd()
            if currDir.count(rootPath) > 1:
                msg = 'The word "%s" occurs multiple times in "%s". The iteration ' \
                      'will stop at the first count.' % (rootPath, currDir)
                logger.warning(msg)
            recursionCount = 10
            while recursionCount and not os.path.split(os.path.split(currDir)[0])[1] == rootPath:
                currDir = os.path.split(currDir)[0]
                recursionCount += -1
            else:
                currDir = os.path.split(currDir)[0]

            if not recursionCount:
                msg = 'The recursion path for find the directory exceeded 10. Check the path' \
                      ' provided.'
                raise Exception(msg)
            self._rootDirectory = currDir

    def pathDirInRoot(self, pathName, create=False, checkIfExists=False):
        r"""

        :param pathName: The directory which is to be linked to the
        :param create:
        :param checkIfExists:
        :return:
        """
        assert not os.path.isabs(pathName), \
            'The path %s is absolute. For a relative path to be created with reference to ' \
            'the root directory, the path provided should be relative.' % pathName

        newPathName = os.path.join(self.rootDirectory, pathName)

        if create and not os.path.exists(newPathName):
            os.mkdir(newPathName)
            msg = "The path %s did not exist. It was created." % (newPathName)
            warnings.warn(msg)
        elif not os.path.exists(newPathName):
            msg = 'The path %s does not currently exist. So any writing operation within ' \
                  'files will fail.' % newPathName
            warnings.warn(msg)

        if checkIfExists:
            assert os.path.exists(newPathName), 'The path "%s" does not exist.' % newPathName
        return newPathName

    def __str__(self):

        pathString = self.directoryName
        return pathString


class UserDirectory(Directory):
    """
    Return default user paths such as desktop, documents etc.

    Typical usecase:
        #Just desktop.

        desktop = UserDirectory().desktop()

        #create a folder inside desktop

        test = UserDirectory().desktop('test')
    """

    def __init__(self):
        """
        This class does not take any inputs as the directory is the home directory of
        the user.

        """

        homeFolder = os.path.expanduser('~')
        self.homeFolder = homeFolder

        Directory.__init__(self, homeFolder)

    def __createPaths(self, folderName, subFolderName=None, overWrite=False):
        """
        This private function does the grind work for other public functions such as
        desktop,documents etc. It is only meant to be called by other functions and not
        directly.

        :param folderName: Name of the default folder.
        :param subFolderName: Name of the subfolder within the default folder that is to
            be created/checked or returned.
        :param overWrite: Boolean. If set to False (default), an exception will be raised
            indicating that the subFolder is not empty. This is to prevent accidental overwrite
            of files from previous operations.
        :return: Folder path that is desired.
        """

        # This is useful for unix-based systems that are path sensitive.
        folderUp = os.path.join(self.homeFolder, folderName.capitalize())
        folderLow = os.path.join(self.homeFolder, folderName)

        if os.path.exists(folderUp):
            folder = folderUp
        elif os.path.exists(folderLow):
            folder = folderLow
        else:
            raise Exception("The path for %s was not found in %s and %s" % \
                            (folderName, folderUp, folderLow))

        # If a subfolder was specified, create that. If the subfolder already exists and
        # is not empty, raise an Exception. The exception can be overridden with the
        # overWrite=True option.
        if subFolderName and isinstance(subFolderName, string_types__):
            returnPath = os.path.join(folder, subFolderName)

            if not os.path.exists(returnPath):
                os.mkdir(returnPath)
                logger.debug('The path %s did not exist and was therefore created' % returnPath)
            elif os.listdir(returnPath) and not overWrite:
                filesInside = '\n'.join(os.listdir(returnPath))
                raise Exception('The folder %s contains the following files:\n %s.\nSet the '
                                '"overWrite" parameter to True to continue writing to this'
                                'file.' % (returnPath, filesInside))
            return returnPath
        else:
            return folder

    def home(self, folderName=None, overWrite=False):
        """

        :param folderName: Name of the folder in the home directory that is to be
            created or returned.
        :param overWrite: Set this to True if it is okay to use the directory even if it
            is not empty.
        :return: folder path.
        :type Directory
        """
        returnPath = self.__createPaths('', subFolderName=folderName, overWrite=overWrite)
        return Directory(returnPath)

    def desktop(self, folderName=None, overWrite=False):
        """

        :param folderName: Name of the folder in the desktop directory that is to be
            created or returned.
        :param overWrite: Set this to True if it is okay to use the directory even if it
            is not empty.
        :return: folder path.
        :type Directory
        """
        returnPath = self.__createPaths('Desktop', subFolderName=folderName, overWrite=overWrite)
        return Directory(returnPath)

    def documents(self, folderName=None, overWrite=False):
        """

        :param folderName: Name of the folder in the documents directory that is to be
            created or returned.
        :param overWrite: Set this to True if it is okay to use the directory even if it
            is not empty.
        :return: folder path.
        :type Directory
        """
        returnPath = self.__createPaths('Documents', subFolderName=folderName, overWrite=overWrite)
        return Directory(returnPath)

    def downloads(self, folderName=None, overWrite=False):
        """

        :param folderName: Name of the folder in the downloads directory that is to be
            created or returned.
        :param overWrite: Set this to True if it is okay to use the directory even if it
            is not empty.
        :return: folder path.
        :type Directory
        """
        returnPath = self.__createPaths('Downloads', subFolderName=folderName, overWrite=overWrite)
        return Directory(returnPath)

    def pictures(self, folderName=None, overWrite=False):
        """

        :param folderName: Name of the folder in the pictures directory that is to be
            created or returned.
        :param overWrite: Set this to True if it is okay to use the directory even if it
            is not empty.
        :return: folder path.
        :type Directory
        """
        returnPath = self.__createPaths('Pictures', subFolderName=folderName, overWrite=overWrite)
        return Directory(returnPath)

    def appdir(self, folderName=None, overWrite=False):
        """
        Return appdata folder in cross-platform fashion. If the folder specific to this repository does not exist
        in the appdata folder, create one.
        :param folderName:
        :param overWrite:
        :return:
        """
        if os.name == 'nt':
            try:
                appData = os.environ['APPDATA']
                appData = os.path.join(appData, 'sarith')
                if not os.path.exists(appData):
                    os.mkdir(appData)
            except KeyError:
                logger.error(
                    "The APPDATA variable was not found in user environment. So a directory called appData%s was "
                    "created in the user directory instead" % "sarith")
                appData = 'appDatasarith'

            returnPath = self.__createPaths(appData, subFolderName=folderName, overWrite=overWrite)
        else:
            assert os.name == 'posix', 'The name of this operation system (%s) is not compatible with this script.' % os.name

            returnPath = self.__createPaths('.sarith', subFolderName=folderName, overWrite=overWrite)

        return Directory(returnPath)


def makeStringPathSafe(inputString, allowNumbers=True,
                       allowFinalDot=False, allowDash=True, allowUnderscore=True,
                       allowDots=False, replaceRemovedChrWith="",assignTempIfNoOutput=True,
                       warnIfOutputStrIsSmall=True):
    u"""
    Take a string that is likely to contain non path safe characters, for example:
     "ad\nbla'{-+\)(ç?" and convert into something that can be used safely for file paths.
     This funciton is being written as a response to some youtube channel names and
     playlists that contain illegal characters. Note that although russian, chinese names
     etc might be path safe, this function will deem them as unsafe. There is way too
     much work invovled in checking for all character sets.

    :param inputString: The string that is to be modified.
    :type inputString: str

    :param allowNumbers: Allow numbers to be present in the path name.
    :param allowFinalDot: Defaults to False, if set to True the final dot in the string
        will be retained. This will likely be useful in case of files where the final
        dot is for extension. Note that this option overrides the allowDots input, that is
        even if dots are not allowed and this one is set to True, the final dot will
        be allowed.
    :param allowDash: Allow "-" in the output. This is not illegal for folders and file
        in windows at least. Defaults to True
    :param allowUnderscore: Allow "_" in the output. This is not illegal for folders and file
        in windows at least. Defautlts to True.
    :param allowDots:Allow "_" in the output. Defautlts to False.
    :param replaceRemovedChrWith: Replace the removed characters with this character. Defaults
        to nothing, but may be changed to "_" or "-" or something like that.
    :param assignTempIfNoOutput: If all the characters are illegal and there is no output
    string, then assign a temp name.
    :param warnIfOutputStrIsSmall: Issue a warning through logger if the output string
         is less than 30% of the input string.
    :return: An output string that has been made path safe as per the given input params
    :rtype: str
    """

    logger.debug("Input string is %s"%inputString)
    inputStringList=list(inputString)

    allowedCharList=[]
    if allowDash:
        allowedCharList.append("-")
    if allowDots:
        allowedCharList.append(".")
    if allowUnderscore:
        allowedCharList.append("_")
    if allowNumbers:
        allowedCharList.extend(map(str,range(10)))
    allowedCharList.extend("abcdefghijklmnopqrstuvwxyz"+"abcdefghijklmnopqrstuvwxyz".upper())

    dotList=[idx for idx,char in enumerate(inputStringList) if char=="."]

    lastDotPosition=dotList[-1] if dotList else None

    outputStringList=[]
    for idx,char in enumerate(inputStringList): #
        if char in allowedCharList:
            outputStringList.append(char)
        elif allowFinalDot and char== "." and idx==lastDotPosition:
            outputStringList.append(".")
        else:
            outputStringList.append(replaceRemovedChrWith)

    if warnIfOutputStrIsSmall and (len("".join(outputStringList))/len(inputStringList)<0.3):
        logger.warning("The outputstring is too small. The input string(%s) had %s "
                       "characters while the output string(%s) has %s charcaters"%(inputString,
                                                                                  len(inputString),
                                                                                  "".join(outputStringList),
                                                                                    len("".join(outputStringList))))

        if not "".join(outputStringList) and assignTempIfNoOutput:
            outputStringList=list(os.path.split(tempfile.mktemp())[-1])
            logger.debug("Since the outputString has no characters and the assignTempIfNoOutput"
                         " has been set to True, the string %s is been provied as the output"%("".join(outputStringList)))




    return "".join(outputStringList)


def removeUnsafeChar(inputString,replaceChar=""):
    """
    Given a string remove potentionally unsafe characters from it so that it can used
    as a filename. This is somewhat less restrictive than makeStringPathSafe which only
    allows alphanumeric characters.
    :param inputString: The string which is to be modified.
    :param replaceChar: The character which is to replace unsafe characters. Defaults to
    nothing.
    :return:
    """
    unsafeChars = ("/", "\\", ">", "<", ":", '"', "'", "|",
                   "*", "§", "$", "%", "&", "(", ")", "{", "}", "*", "+", "#", "°", "~",
                   "€", "@",
                   ";", "µ", "^", " ")
    inputStringList=list(inputString)
    outputString=""

    for idx,char in enumerate(inputStringList):
        if char in unsafeChars:
            outputString+=replaceChar
        else:
            outputString+=char

    logger.debug("Replaced the input string (%s) with (%s)"%(inputString,outputString))

    return outputString





if __name__ == "__main__":
    pass