"""
Functions that open a file, read the lines and return it as a list. Currently the following
functions have been defined:
    returnLines: Given a filePath read it and return a list of lines. The args and kwargs can be used
    to specify further options for the open command.
"""

import logging

logger=logging.getLogger("__main__")

def returnLines(filePath,ignoreEmptyLines=True,stripLines=True,*args,**kwargs):
    """

    Given a filePath read it and return a list of lines. The args and kwargs can be used
    to specify further options for the open command.

    :param filePath: This is the absolute path of the file to be opened.
    :type filePath: str

    :param ignoreEmptyLines: If set to true, lines with only spaces, tabs etc and nothing
        else will be ignored and not included in the final list.
    :type ignoreEmptyLines: bool

    :param stripLines: If set to True, then the returned lines will be stripped using .strip()
    :type stripLines: bool

    :param args: These will be passed on to the open() command and can be used to specify
        things like encoding.

    :param kwargs: These will be passed on to the open() command and can be used to specify
        things like encoding.

    :return: A list containing lines from the file that was read.
    """
    lineList=[]
    logger.debug("Opening the file %s for reading"%filePath)

    with open(filePath,*args,**kwargs) as fileData:
        for idx,lines in enumerate(fileData):
            if ignoreEmptyLines and not lines.strip():
                continue
            elif not lines.strip():
                if stripLines:
                    lineList.append(lines.strip())
                else:
                    lineList.append(lines)
            else:
                if stripLines:
                    lineClean=lines.strip()
                else:
                    lineClean=lines

                lineList.append(lineClean)
    if lineList:
        logger.debug("The file had %s lines. Based on the filtering criteria the list has %s lines"%(idx+1,len(lineList)))
    else:
        logger.debug("No lines were read in this file")

    return lineList


if __name__=="__main__":
    pass

