"""Contains a few functions for writing strings, normal or binary into files. Has a whole bunch of testing, and additional
 functionlity built into it. Presently, three functions are available:

 + writeToFile: Write a string to file.
 + appendToFile: Append a string to file.
 + writeToFileBinary: Write byte-strings to a file.
 """

import os
import logging
import locale
logger=logging.getLogger("__main__")

def __writerTemplate(string, fileName, mode,log=True, encoding=None, errors=None, newLine=None,):
    r"""
    This is the general template function for writing in different modes. This function should not be called on it's own
    and is meant to be invoked by other functions with limited options instead.

    :param string: The string that is to be written to file. Can be a byte-string or a regular string.
    :param fileName: The file path to which the string is to be written to.
    :param mode: The writing mode. Valid modes are: 'w':writing ,'a':appending ,'w+b':binary writing ,'w+':writing ,'wb':binary writing.
        If the mode is any other than these, then an exception will be raised through an assert statement.
    :param log: If set to True, then different stages of the write operation will be logged.
    :param encoding: Encoding option for writing. Since the default is None, the open function will fetch the value from
         locale.getpreferredencoding(). Other plausible options are: utf-8,utf8,latin-1,latin1,iso-8859-1,iso8859-1,
         mbcs,ascii,us-ascii,utf-16,utf16,utf-32,utf32,utf_8,latin_1,iso_8859_1,iso8859_1,us_ascii,utf_16,utf_32
    :param errors: errors is an optional string that specifies how encoding and decoding errors are to be handled—this
        cannot be used in binary mode. A variety of standard error handlers are available (listed under Error Handlers),
        though any error handling name that has been registered with codecs.register_error() is also valid.
    :param newLine: Controls how universal newlines mode works (it only applies to text mode). It can be None,
        '', '\n', '\r', and '\r\n'.
    :return:
    """

    prefEncoding=locale.getpreferredencoding()

    otherEncodings=["utf-8", "utf8", "latin-1", "latin1", "iso-8859-1", "iso8859-1", "mbcs", "ascii",
                    "us-ascii", "utf-16", "utf16", "utf-32", "utf32",'utf_8', 'latin_1', 'iso_8859_1',
                    'iso8859_1', 'us_ascii', 'utf_16', 'utf_32']


    possibleEncodings=[prefEncoding]+otherEncodings

    possibleModes={"w":"writing","a":"appending","w+b":"binary writing","w+":"writing","wb":"binary writing"}

    assert mode in possibleModes,"The specified mode '%s' is not valid. Only the following types are valid: %s"%(mode,",".join(["'%s':%s "%(key,val) for key,val in possibleModes.items()]))


    if log:

        if type(string) is bytes:
            stringType=" byte "
        else:
            stringType=" "
        logger.debug("Writing a string%sof length %s to the file %s"%(stringType,len(string),fileName))

        if encoding and not (encoding in possibleEncodings):
            logger.warning("The specified encoding type '%s' is not within the possible list of encodings. "
                           "The write operation might not be successful"%encoding)

        fileExists=os.path.exists(fileName)

        if fileExists:
            writeStatus=" and will be overwritten." if not ("a" in mode) else "and will be appended to."
            logger.debug("The file '%s' already exists %s"%(fileName,writeStatus))
        else:
            logger.debug("The file '%s' does not exist"%fileName)


    if 'b' in mode:
        assert type(string) is bytes,"The input for 'string' is not a byte and is %s instead, however the writing mode is '%s', so writing cannot proceed."%(type(string),mode)
        assert encoding is None,"No encoding should be specified for binary mode."
        assert newLine is None,"No newline argument (%s) is allowed for writing in binary mode"%newLine
        fileToWrite=open(fileName,mode,errors=errors,newline=newLine,)
    else:
        assert type(string) is str, "The input for 'string' is not a 'str' and is %s instead, however the writing mode is '%s', so writing cannot proceed." % (type(string), mode)
        fileToWrite = open(fileName, mode, encoding=encoding, errors=errors, newline=newLine)

    with fileToWrite:
        fileToWrite.write(string)

        if log:
            mode="mode: '%s'(%s)"%(fileToWrite.mode,possibleModes[fileToWrite.mode])
            try:
                encoding,errors,newLine=", encoding: '%s',"%fileToWrite.encoding,"errors: '%s', "%fileToWrite.errors,"newLines: '%s'."%fileToWrite.newlines
            except AttributeError:
                encoding,errors,newLine="","",""
            logger.debug("The write operation used following attributes: %s%s%s%s"%(mode,encoding,errors,newLine))

    logger.debug("Finished writing, the 'closed' status of the file descriptor is '%s'. Exiting" % fileToWrite.closed)

    return fileName

def writeToFile(string:str,fileName:str,log=True, encoding:str=None, errors=None, newLine:str=None) -> str:
    r"""
    Write a string to file and return the filename in case the operation was successful.

    :param string: The string that is to be written to file. Can be a byte-string or a regular string.
    :param fileName: The file path to which the string is to be written to.
    :param log: If set to True, then different stages of the write operation will be logged.
    :param encoding: Encoding option for writing. Since the default is None, the open function will fetch the value from
         locale.getpreferredencoding(). Other plausible options are: utf-8,utf8,latin-1,latin1,iso-8859-1,iso8859-1,
         mbcs,ascii,us-ascii,utf-16,utf16,utf-32,utf32,utf_8,latin_1,iso_8859_1,iso8859_1,us_ascii,utf_16,utf_32
    :param errors: errors is an optional string that specifies how encoding and decoding errors are to be handled—this
        cannot be used in binary mode. A variety of standard error handlers are available (listed under Error Handlers),
        though any error handling name that has been registered with codecs.register_error() is also valid.
    :param newLine: Controls how universal newlines mode works (it only applies to text mode). It can be None,
        '', '\n', '\r', and '\r\n'.
    :return: Filename that the string was written to if the operation was successful.
    """
    return __writerTemplate(string=string,fileName=fileName,mode="w",log=log,encoding=encoding,errors=errors,newLine=newLine)

def appendToFile(string:str,fileName:str,log=True, encoding=None, errors=None, newLine=None) -> str:
    r"""
    Append a string to file and return the filename in case the operation was successful.

    :param string: The string that is to be written to file. Can be a byte-string or a regular string.
    :param fileName: The file path to which the string is to be written to.
    :param log: If set to True, then different stages of the write operation will be logged.
    :param encoding: Encoding option for writing. Since the default is None, the open function will fetch the value from
         locale.getpreferredencoding(). Other plausible options are: utf-8,utf8,latin-1,latin1,iso-8859-1,iso8859-1,
         mbcs,ascii,us-ascii,utf-16,utf16,utf-32,utf32,utf_8,latin_1,iso_8859_1,iso8859_1,us_ascii,utf_16,utf_32
    :param errors: errors is an optional string that specifies how encoding and decoding errors are to be handled—this
        cannot be used in binary mode. A variety of standard error handlers are available (listed under Error Handlers),
        though any error handling name that has been registered with codecs.register_error() is also valid.
    :param newLine: Controls how universal newlines mode works (it only applies to text mode). It can be None,
        '', '\n', '\r', and '\r\n'.
    :return: Filename that the string was written to if the operation was successful.
    """
    return __writerTemplate(string=string,fileName=fileName,mode="a",log=log,encoding=encoding,errors=errors,newLine=newLine)

def writeToFileBinary(string:bytes,fileName:str,log=True) -> str:
    r"""
    Write byte-strings to a file and return the filename in case the operation was successful.

    :param string: The string that is to be written to file. Can be a byte-string or a regular string.
    :param fileName: The file path to which the string is to be written to.
    :param log: If set to True, then different stages of the write operation will be logged.
    :return: Filename that the string was written to if the operation was successful.
    """
    return __writerTemplate(string=string,fileName=fileName,mode="wb",log=log)

if __name__ == "__main__":
    pass